"""
Module that implements asyncio.coroutine function wrappers to be used
by param internal callbacks. These are defined in a separate file due
to py2 incompatibility with both `async/await` and `yield from` syntax.
"""

# import asyncio

# def generate_depends(func):
#     @asyncio.coroutine
#     def _depends(*args, **kw):
#         yield from func(*args, **kw) # noqa: E999
#     return _depends



# def generate_callback(func, dependencies, kw):
#     @asyncio.coroutine
#     def cb(*events):
#         args = (getattr(dep.owner, dep.name) for dep in dependencies)
#         dep_kwargs = {n: getattr(dep.owner, dep.name) for n, dep in kw.items()}
#         yield from func(*args, **dep_kwargs) # noqa: E999
#     return cb



import typing
import jsonschema
from enum import Enum
from types import FunctionType, MethodType
from inspect import iscoroutinefunction, getfullargspec

from ..param.parameterized import ParameterizedFunction
from .utils import issubklass, pep8_to_URL_path, isclassmethod
from .dataklasses import ActionInfoValidator
from .constants import USE_OBJECT_NAME, UNSPECIFIED, HTTP_METHODS, JSON
from .config import global_config



__action_kw_arguments__ = ['safe', 'idempotent', 'synchronous'] 
   
def action(URL_path : str = USE_OBJECT_NAME, http_method : str = HTTP_METHODS.POST, 
            state : typing.Optional[typing.Union[str, Enum]] = None, input_schema : typing.Optional[JSON] = None,
            output_schema : typing.Optional[JSON] = None, create_task : bool = False, **kwargs) -> typing.Callable:
    """
    Use this function as a decorate on your methods to make them accessible remotely. For WoT, an action affordance schema 
    for the method is generated.
    
    Parameters
    ----------
    URL_path: str, optional 
        The path of URL under which the object is accessible. defaults to name of the object.
    http_method: str, optional
        HTTP method (GET, POST, PUT etc.). defaults to POST.
    state: str | Tuple[str], optional 
        state machine state under which the object can executed. When not provided,
        the action can be executed under any state.
    input_schema: JSON 
        schema for arguments to validate them.
    output_schema: JSON 
        schema for return value, currently only used to inform clients which is supposed to validate on its won. 
    **kwargs:
        safe: bool 
            indicate in thing description if action is safe to execute 
        idempotent: bool 
            indicate in thing description if action is idempotent (for example, allows HTTP client to cache return value)
        synchronous: bool
            indicate in thing description if action is synchronous (not long running)
    Returns
    -------
    Callable
        returns the callable object as it is
    """
    
    def inner(obj):
        original = obj
        if (not isinstance(obj, (FunctionType, MethodType)) and not isclassmethod(obj) and 
            not issubklass(obj, ParameterizedFunction)):
                raise TypeError(f"target for action or is not a function/method. Given type {type(obj)}") from None 
        if isclassmethod(obj):
            obj = obj.__func__
        if obj.__name__.startswith('__'):
            raise ValueError(f"dunder objects cannot become remote : {obj.__name__}")
        if hasattr(obj, '_remote_info') and not isinstance(obj._remote_info, ActionInfoValidator): 
            raise NameError(
                "variable name '_remote_info' reserved for hololinked package. ",  
                "Please do not assign this variable to any other object except hololinked.server.dataklasses.ActionInfoValidator."
            )             
        else:
            obj._remote_info = ActionInfoValidator() 
        obj_name = obj.__qualname__.split('.')
        if len(obj_name) > 1: # i.e. its a bound method, used by Thing
            if URL_path == USE_OBJECT_NAME: 
                obj._remote_info.URL_path = f'/{pep8_to_URL_path(obj_name[1])}'
            else:
                if not URL_path.startswith('/'):
                    raise ValueError(f"URL_path should start with '/', please add '/' before '{URL_path}'")
                obj._remote_info.URL_path = URL_path
            obj._remote_info.obj_name = obj_name[1] 
        elif len(obj_name) == 1 and isinstance(obj, FunctionType):  # normal unbound function - used by HTTPServer instance
            if URL_path is USE_OBJECT_NAME:
                obj._remote_info.URL_path = f'/{pep8_to_URL_path(obj_name[0])}'
            else:
                if not URL_path.startswith('/'):
                    raise ValueError(f"URL_path should start with '/', please add '/' before '{URL_path}'")
                obj._remote_info.URL_path = URL_path
            obj._remote_info.obj_name = obj_name[0] 
        else:
            raise RuntimeError(f"Undealt option for decorating {obj} or decorators wrongly used")
        if http_method is not UNSPECIFIED:  
            if isinstance(http_method, str):
                obj._remote_info.http_method = (http_method,)
            else:
                obj._remote_info.http_method = http_method 
        if state is not None:
            if isinstance(state, (Enum, str)):
                obj._remote_info.state = (state,)
            else:
                obj._remote_info.state = state     
        if 'request' in getfullargspec(obj).kwonlyargs:
            obj._remote_info.request_as_argument = True
        obj._remote_info.isaction = True
        obj._remote_info.argument_schema = input_schema
        obj._remote_info.return_value_schema = output_schema
        obj._remote_info.obj = original
        obj._remote_info.create_task = create_task
        obj._remote_info.safe = kwargs.get('safe', False)
        obj._remote_info.idempotent = kwargs.get('idempotent', False)
        obj._remote_info.synchronous = kwargs.get('synchronous', False)
        
        if issubklass(obj, ParameterizedFunction):
            obj._remote_info.iscoroutine = iscoroutinefunction(obj.__call__)
            obj._remote_info.isparameterized = True
        else:
            obj._remote_info.iscoroutine = iscoroutinefunction(obj)
            obj._remote_info.isparameterized = False 
        if global_config.validate_schemas and input_schema:
            jsonschema.Draft7Validator.check_schema(input_schema)
        if global_config.validate_schemas and output_schema:
            jsonschema.Draft7Validator.check_schema(output_schema)
        
        return original
    if callable(URL_path):
        raise TypeError("URL_path should be a string, not a function/method, did you decorate your action wrongly?")
    if any(key not in __action_kw_arguments__ for key in kwargs.keys()):
        raise ValueError("Only 'safe', 'idempotent', 'synchronous' are allowed as keyword arguments, " + 
                        f"unknown arguments found {kwargs.keys()}")
    return inner 



__all__ = [
    action.__name__
]





from typing import Dict, List, Any, Union
from dataclasses import dataclass, field, asdict

from .constants import HTTP_METHODS
from .serializers import JSONSerializer


@dataclass
class postman_collection:
    """
    Generate postman JSON of schema v2.1.0 (https://schema.postman.com/collection/json/v2.1.0/draft-04/collection.json)
    """
    info : "postman_collection_info"
    item : Union[List[Union["postman_item", "postman_itemgroup"]], Dict[str, Any]]

    def add_item(self, item : Union["postman_item", "postman_itemgroup"]) -> None:
        if isinstance(self.item, dict):
            raise ValueError("Please define item as a list before adding requests to item.")
        self.item.append(item)

    def json(self):
        return asdict(self)
    
    def save_json_file(self, filename = 'collection.json'):
        with open(filename, 'w') as file: 
            JSONSerializer.generic_dump(self.json(), file)

    @classmethod
    def build(cls, instance, domain_prefix : str) -> Dict[str, Any]:
        from .thing import Thing
        from .dataklasses import HTTPResource, RemoteResource
        assert isinstance(instance, Thing) # type definition
        try:
            return instance._postman_collection
        except AttributeError:
            pass 
        properties_folder = postman_itemgroup(name='properties')
        methods_folder = postman_itemgroup(name='methods')
        events_folder = postman_itemgroup(name='events')

        collection = postman_collection(
            info = postman_collection_info(
                name = instance.__class__.__name__,
                description = "API endpoints available for Thing", 
            ),
            item = [ 
                properties_folder,
                methods_folder                
            ]
        )

        for http_method, resource in instance.httpserver_resources.items():
            # i.e. this information is generated only on the httpserver accessible resrouces...
            for URL_path, httpserver_data in resource.items():
                if isinstance(httpserver_data, HTTPResource):
                    scada_info : RemoteResource
                    try:
                        scada_info = instance.instance_resources[httpserver_data.instruction]
                    except KeyError:
                        property_path_without_RW = httpserver_data.instruction.rsplit('/', 1)[0]
                        scada_info = instance.instance_resources[property_path_without_RW]
                    item = postman_item(
                        name = scada_info.obj_name,
                        request = postman_http_request(
                            description=scada_info.obj.__doc__,
                            url=domain_prefix + URL_path, 
                            method=http_method,
                        )
                    )
                    if scada_info.isproperty:
                        properties_folder.add_item(item)
                    elif scada_info.iscallable:
                        methods_folder.add_item(item)
        
        instance._postman_collection = collection
        return collection


@dataclass
class postman_collection_info:
    """
    info field of postman collection 
    """
    name : str 
    description : str
    version : str = field(default="v2.1.0")
    schema : str = field(default="https://schema.getpostman.com/json/collection/v2.1.0/")

    def json(self):
        return asdict(self)

@dataclass
class postman_item:
    """
    item field of postman collection 
    """
    name : str 
    request : "postman_http_request"    
    description : Union[str, None] = field(default=None)

    def json(self):
        return asdict(self)

@dataclass
class postman_http_request:
    """
    HTTP request item of postman collection
    """
    url : str 
    header : Union[List[Dict[str, Any]], None] = field(default=None)  
    body : Union[Dict[str, Any], None] = field(default=None)
    method : str = field(default=HTTP_METHODS.POST) 
    description : Union[str, None] = field(default=None)

    def json(self):
        json_dict = asdict(self)
        if self.header is None:
            json_dict.pop("header", None)
        if self.body is None:
            json_dict.pop("body", None)
        return json_dict

@dataclass
class postman_itemgroup:
    """
    item group of postman collection
    """
    name : str 
    item : Union[postman_item, "postman_itemgroup", 
                List[Union["postman_item", "postman_itemgroup"]]] = field(default_factory=list)
    description : Union[str, None] = field(default=None)

    def add_item(self, item : Union["postman_item", "postman_itemgroup"]) -> None:
        if isinstance(self.item, dict):
            raise ValueError("Please define item as a list before adding requests to item.")
        if isinstance(self.item, list):
            self.item.append(item)
        else: 
            raise ValueError(f"itemgroup must be list, not type {type(item)}")
        
    def json(self):
        return asdict(self)
    


__all__ = [
    'postman_collection',
    'postman_collection_info',
    'postman_item',
    'postman_http_request',
    'postman_itemgroup'
]


import zmq
import logging
import typing
from enum import StrEnum, IntEnum, Enum


# types
JSONSerializable = typing.Union[typing.Dict[str, typing.Any], list, str, int, float, None]
JSON = typing.Dict[str, JSONSerializable]

# decorator constants 
# naming
USE_OBJECT_NAME : str = "USE_OBJECT_NAME"
# state machine
ANY_STATE   : str = "ANY_STATE"
UNSPECIFIED : str = "UNSPECIFIED"
# types


class ResourceTypes(StrEnum):
    """
    Exposed resource types. This repository does not rely on python's internal getattr, setattr and type checking, 
    to expose RPC operations, but rather segrates resource types and operates on them in predefined ways. This is to allow
    addition or removal of resource types which may demand more integration with HTTP.  
    """
    PROPERTY = "PROPERTY"
    ACTION = "ACTION"
    EVENT = "EVENT"
    IMAGE_STREAM = "IMAGE_STREAM"
    FILE = "FILE"


class CommonRPC(StrEnum):
    """some common RPC and their associated instructions for quick access by lower level code"""

    ZMQ_RESOURCES = '/resources/zmq-object-proxy'
    HTTP_RESOURCES = '/resources/http-server'
    OBJECT_INFO = '/object-info'
    PING = '/ping'

    @classmethod
    def zmq_resource_read(cls, instance_name : str) -> str:
        return f"/{instance_name}{cls.ZMQ_RESOURCES}/read"

    @classmethod
    def http_resource_read(cls, instance_name : str) -> str:
        return f"/{instance_name}{cls.HTTP_RESOURCES}/read"
    
    @classmethod
    def object_info_read(cls, instance_name : str) -> str: 
        return f"/{instance_name}{cls.OBJECT_INFO}/read"
    
    @classmethod
    def object_info_write(cls, instance_name : str) -> str: 
        return f"/{instance_name}{cls.OBJECT_INFO}/write"


class REGEX(StrEnum):
    """common regexes"""

    states = '[A-Za-z_]+[A-Za-z_ 0-9]*'
    url = r'[\-a-zA-Z0-9@:%._\/\+~#=]{1,256}' 


class HTTP_METHODS(StrEnum):
    """currently supported HTTP request methods"""

    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    PATCH = 'PATCH'
    # OPTIONS = 'OPTIONS'

http_methods = [member for member in HTTP_METHODS._member_map_]


# Logging 
class LOGLEVEL(IntEnum):
    """``logging.Logger`` log levels"""
    DEBUG    = logging.DEBUG
    INFO     = logging.INFO
    CRITICAL = logging.CRITICAL
    ERROR    = logging.ERROR
    WARN     = logging.WARN
    FATAL    = logging.FATAL


# ZMQ
class ZMQ_PROTOCOLS(StrEnum):
    """
    supported ZMQ transport protocols - TCP, IPC, INPROC

    TCP - needs socket address additional specified like tcp://0.0.0.0:{port}
    IPC - within python, multiprocess applications, can be autogenerated
    INPROC - within python, multithreaded applications, can be autogenerated
    """
    TCP = "TCP"
    IPC = "IPC"
    INPROC = "INPROC"


class ClientMessage(IntEnum):
    """
    ZMQ client sent message indexing for accessing message indices with names 
    instead of numbers
    """
    ADDRESS = 0
    CLIENT_TYPE = 2
    MESSAGE_TYPE = 3
    MESSAGE_ID = 4
    TIMEOUT = 5
    INSTRUCTION = 6
    ARGUMENTS = 7
    EXECUTION_CONTEXT = 8


class ServerMessage(IntEnum):
    """
    ZMQ server sent message indexing for accessing message indices with names 
    instead of numbers
    """
    ADDRESS = 0
    SERVER_TYPE = 2
    MESSAGE_TYPE = 3
    MESSAGE_ID = 4
    DATA = 5
    ENCODED_DATA = 6


class ServerTypes(Enum):
    "type of ZMQ servers"

    UNKNOWN_TYPE = b'UNKNOWN' 
    EVENTLOOP = b'EVENTLOOP'  
    THING = b'THING'
    POOL = b'POOL'
       
       
class ClientTypes(Enum):
    "type of ZMQ clients"

    HTTP_SERVER = b'HTTP_SERVER'
    PROXY = b'PROXY'
    TUNNELER = b'TUNNELER' # message passer from inproc client to inrproc server within RPC


class HTTPServerTypes(StrEnum):
    "types of HTTP server"

    SYSTEM_HOST = 'SYSTEM_HOST'
    THING_SERVER = 'THING_SERVER'


class Serializers(StrEnum):
    """
    allowed serializers

    - PICKLE : pickle
    - JSON : msgspec.json
    - SERPENT : serpent
    - MSGPACK : msgspec.msgpack
    """
    PICKLE = 'pickle'
    JSON = 'json'
    SERPENT = 'serpent'
    MSGPACK = 'msgpack'


class ZMQSocketType(IntEnum):
    PAIR = zmq.PAIR
    PUB = zmq.PUB
    SUB = zmq.SUB
    REQ = zmq.REQ
    REP = zmq.REP
    DEALER = zmq.DEALER
    ROUTER = zmq.ROUTER
    PULL = zmq.PULL
    PUSH = zmq.PUSH
    XPUB = zmq.XPUB
    XSUB = zmq.XSUB
    STREAM = zmq.STREAM
    # Add more socket types as needed


ZMQ_EVENT_MAP = {}
for name in dir(zmq):
    if name.startswith('EVENT_'):
        value = getattr(zmq, name)
        ZMQ_EVENT_MAP[value] = name




__all__ = [
    Serializers.__name__, 
    HTTP_METHODS.__name__,
    ZMQ_PROTOCOLS.__name__
]


import typing 
import inspect
from types import FunctionType, MethodType
from enum import EnumMeta, Enum, StrEnum

from ..param.parameterized import Parameterized, edit_constant
from .utils import getattr_without_descriptor_read
from .exceptions import StateMachineError
from .dataklasses import RemoteResourceInfoValidator
from .property import Property
from .properties import ClassSelector, TypedDict, Boolean




class StateMachine:
    """
    A container class for state machine related information. Each ``Thing`` class can only have one state machine  
    instantiated in a reserved class-level attribute ``state_machine``. The ``state`` attribute defined at the ``Thing``
    can be subscribed for state change events from this state machine. 
    """
    initial_state = ClassSelector(default=None, allow_None=True, constant=True, class_=(Enum, str), 
                        doc="initial state of the machine") # type: typing.Union[Enum, str]
    states = ClassSelector(default=None, allow_None=True, constant=True, class_=(EnumMeta, tuple, list),
                        doc="list/enum of allowed states") # type: typing.Union[EnumMeta, tuple, list]
    on_enter = TypedDict(default=None, allow_None=True, key_type=str,
                        doc="""callbacks to execute when a certain state is entered; 
                        specfied as map with state as keys and callbacks as list""") # typing.Dict[str, typing.List[typing.Callable]]
    on_exit = TypedDict(default=None, allow_None=True, key_type=str,
                        doc="""callbacks to execute when certain state is exited; 
                        specfied as map with state as keys and callbacks as list""") # typing.Dict[str, typing.List[typing.Callable]]
    machine = TypedDict(default=None, allow_None=True, item_type=(list, tuple), key_type=str, # i.e. its like JSON
                        doc="the machine specification with state as key and objects as list") # typing.Dict[str, typing.List[typing.Callable, Property]]
    valid = Boolean(default=False, readonly=True, fget=lambda self: self._valid, 
                        doc="internally computed, True if states, initial_states and the machine is valid")
    
    def __init__(self, 
            states : typing.Union[EnumMeta, typing.List[str], typing.Tuple[str]], *, 
            initial_state : typing.Union[StrEnum, str], push_state_change_event : bool = True,
            on_enter : typing.Dict[str, typing.Union[typing.List[typing.Callable], typing.Callable]] = {}, 
            on_exit  : typing.Dict[str, typing.Union[typing.List[typing.Callable], typing.Callable]] = {}, 
            **machine : typing.Dict[str, typing.Union[typing.Callable, Property]]
        ) -> None:
        """
        Parameters
        ----------
        states: Enum
            enumeration of states 
        initial_state: str 
            initial state of machine 
        push_state_change_event : bool, default True
            when the state changes, an event is pushed with the new state
        on_enter: Dict[str, Callable | Property] 
            callbacks to be invoked when a certain state is entered. It is to be specified 
            as a dictionary with the states being the keys
        on_exit: Dict[str, Callable | Property]
            callbacks to be invoked when a certain state is exited. 
            It is to be specified as a dictionary with the states being the keys
        **machine:
            state name: List[Callable, Property]
                directly pass the state name as an argument along with the methods/properties which are allowed to execute 
                in that state
        """
        self._valid = False
        self.on_enter = on_enter
        self.on_exit  = on_exit
        # None cannot be passed in, but constant is necessary. 
        self.states   = states
        self.initial_state = initial_state
        self.machine = machine
        self.push_state_change_event = push_state_change_event
        # if :
        #     self.state_change_event = Event('state-change') 

    def _prepare(self, owner : Parameterized) -> None:
        if self.states is None and self.initial_state is None:    
            self._valid = False 
            self._state = None
            return
        elif self.initial_state not in self.states:
            raise AttributeError(f"specified initial state {self.initial_state} not in Enum of states {self.states}.")

        self._state = self._get_machine_compliant_state(self.initial_state)
        self.owner = owner
        owner_properties = owner.parameters.descriptors.values() # same as owner.properties.descriptors.values()
        owner_methods = [obj[0] for obj in inspect._getmembers(owner, inspect.ismethod, getattr_without_descriptor_read)]
        
        if isinstance(self.states, list):
            self.__class__.states.constant = False 
            self.states = tuple(self.states) # freeze the list of states
            self.__class__.states.constant = True   
            
        # first validate machine
        for state, objects in self.machine.items():
            if state in self:
                for resource in objects:
                    if hasattr(resource, '_remote_info'):
                        assert isinstance(resource._remote_info, RemoteResourceInfoValidator) # type definition
                        if resource._remote_info.isaction and resource._remote_info.obj_name not in owner_methods: 
                            raise AttributeError("Given object {} for state machine does not belong to class {}".format(
                                                                                                resource, owner))
                        if resource._remote_info.isproperty and resource not in owner_properties: 
                            raise AttributeError("Given object {} - {} for state machine does not belong to class {}".format(
                                                                                                resource.name, resource, owner))
                        if resource._remote_info.state is None: 
                            resource._remote_info.state = self._get_machine_compliant_state(state)
                        else: 
                            resource._remote_info.state = resource._remote_info.state + (self._get_machine_compliant_state(state), ) 
                    else: 
                        raise AttributeError(f"Object {resource} was not made remotely accessible," + 
                                    " use state machine with properties and actions only.")
            else:
                raise StateMachineError("Given state {} not in states Enum {}".format(state, self.states.__members__))
            
        # then the callbacks 
        for state, objects in self.on_enter.items():
            if isinstance(objects, list):
                self.on_enter[state] = tuple(objects) 
            elif not isinstance(objects, (list, tuple)):
                self.on_enter[state] = (objects, )
            for obj in self.on_enter[state]: # type: ignore
                if not isinstance(obj, (FunctionType, MethodType)):
                    raise TypeError(f"on_enter accept only methods. Given type {type(obj)}.")

        for state, objects in self.on_exit.items():
            if isinstance(objects, list):
                self.on_exit[state] = tuple(objects) # type: ignore
            elif not isinstance(objects, (list, tuple)):
                self.on_exit[state] = (objects, ) # type: ignore
            for obj in self.on_exit[state]: # type: ignore
                if not isinstance(obj, (FunctionType, MethodType)):
                    raise TypeError(f"on_enter accept only methods. Given type {type(obj)}.")     
        self._valid = True
        
    def __contains__(self, state : typing.Union[str, StrEnum]):
        if isinstance(self.states, EnumMeta) and state in self.states.__members__:
            return True
        elif isinstance(self.states, tuple) and state in self.states:
            return True
        return False
        
    def _get_machine_compliant_state(self, state) -> typing.Union[StrEnum, str]:
        """
        In case of not using StrEnum or iterable of str, 
        this maps the enum of state to the state name.
        """
        if isinstance(state, str):
            return state 
        if isinstance(state, Enum):
            return state.name
        raise TypeError(f"cannot comply state to a string : {state} which is of type {type(state)}.")
    

    def get_state(self) -> typing.Union[str, StrEnum, None]:
        """
        return the current state. one can also access the property `current state`.
        
        Returns
        -------
        current state: str
        """
        return self._state
        
    def set_state(self, value : typing.Union[str, StrEnum, Enum], push_event : bool = True, 
                skip_callbacks : bool = False) -> None:
        """ 
        set state of state machine. Also triggers state change callbacks if skip_callbacks=False and pushes a state 
        change event when push_event=True. One can also set state using '=' operator of `current_state` property in which case 
        callbacks will be called. If originally an enumeration for the list of allowed states was supplied, 
        then an enumeration member must be used to set state. If a list of strings were supplied, then a string is accepted. 

        Raises
        ------
        ValueError: 
            if the state is not found in the allowed states
        """
    
        if value in self.states:
            previous_state = self._state
            self._state = self._get_machine_compliant_state(value)
            if push_event and self.push_state_change_event and hasattr(self.owner, 'event_publisher'):
                self.owner.state # just acces to trigger the observable event
            if skip_callbacks:
                return 
            if previous_state in self.on_exit:
                for func in self.on_exit[previous_state]:
                    func(self.owner)
            if self._state in self.on_enter:
                for func in self.on_enter[self._state]: 
                    func(self.owner)
        else:   
            raise ValueError("given state '{}' not in set of allowed states : {}.".format(value, self.states))
                
    current_state = property(get_state, set_state, None, 
        doc = """read and write current state of the state machine""")

    def has_object(self, object : typing.Union[Property, typing.Callable]) -> bool:
        """
        returns True if specified object is found in any of the state machine states. 
        Supply unbound method for checking methods, as state machine is specified at class level
        when the methods are unbound. 
        """
        for objects in self.machine.values():
            if object in objects:
                return True 
        return False
    




import typing 
import threading
import jsonschema

from ..param.parameterized import Parameterized, ParameterizedMetaclass
from .constants import JSON 
from .utils import pep8_to_URL_path
from .config import global_config
from .zmq_message_brokers import EventPublisher
from .dataklasses import ServerSentEvent
from .security_definitions import BaseSecurityDefinition



class Event:
    """
    Asynchronously push arbitrary messages to clients. Apart from default events created by the package (like state
    change event, observable properties etc.), events are supposed to be created at class level or at ``__init__`` 
    as a instance attribute, otherwise their publishing socket is unbound and will lead to ``AttributeError``.  

    Parameters
    ----------
    name: str
        name of the event, specified name may contain dashes and can be used on client side to subscribe to this event.
    URL_path: str
        URL path of the event if a HTTP server is used. only GET HTTP methods are supported. 
    doc: str
        docstring for the event
    schema: JSON
        schema of the event, if the event is JSON complaint. HTTP clients can validate the data with this schema. There
        is no validation on server side.
    """
    # security: Any
    #     security necessary to access this event.

    __slots__ = ['friendly_name', '_internal_name', '_obj_name',
                'doc', 'schema', 'URL_path', 'security', 'label', 'owner']


    def __init__(self, friendly_name : str, URL_path : typing.Optional[str] = None, doc : typing.Optional[str] = None, 
                schema : typing.Optional[JSON] = None, # security : typing.Optional[BaseSecurityDefinition] = None,
                label : typing.Optional[str] = None) -> None:
        self.friendly_name = friendly_name 
        self.doc = doc 
        if global_config.validate_schemas and schema:
            jsonschema.Draft7Validator.check_schema(schema)
        self.schema = schema
        self.URL_path = URL_path or f'/{pep8_to_URL_path(friendly_name)}'
        # self.security = security
        self.label = label
       

    def __set_name__(self, owner : ParameterizedMetaclass, name : str) -> None:
        self._internal_name = f"{pep8_to_URL_path(name)}-dispatcher"
        self._obj_name = name
        self.owner = owner

    @typing.overload
    def __get__(self, obj, objtype) -> "EventDispatcher":
        ...

    def __get__(self, obj : ParameterizedMetaclass, objtype : typing.Optional[type] = None):
        try:
            if not obj:
                return self
            return obj.__dict__[self._internal_name]
        except KeyError:
            raise AttributeError("Event object not yet initialized, please dont access now." +
                                " Access after Thing is running.")
        
    def __set__(self, obj : Parameterized, value : typing.Any) -> None:
        if isinstance(value, EventDispatcher):
            value._remote_info.name = self.friendly_name
            value._remote_info.obj_name = self._obj_name
            value._owner_inst = obj
            current_obj = obj.__dict__.get(self._internal_name, None) # type: typing.Optional[EventDispatcher]
            if current_obj and current_obj._publisher:
                current_obj._publisher.unregister(current_obj)
            obj.__dict__[self._internal_name] = value 
        else:
            raise TypeError(f"Supply EventDispatcher object to event {self._obj_name}, not type {type(value)}.")

    

class EventDispatcher:
    """
    The actual worker which pushes the event. The separation is necessary between ``Event`` and 
    ``EventDispatcher`` to allow class level definitions of the ``Event`` 
    """
    def __init__(self, unique_identifier : str) -> None:
        self._unique_identifier = bytes(unique_identifier, encoding='utf-8')         
        self._publisher = None
        self._remote_info = ServerSentEvent(unique_identifier=unique_identifier)
        self._owner_inst = None

    @property
    def publisher(self) -> "EventPublisher": 
        """
        Event publishing PUB socket owning object.
        """
        return self._publisher
    
    @publisher.setter
    def publisher(self, value : "EventPublisher") -> None:
        if not self._publisher:
            self._publisher = value
            self._publisher.register(self)
        else:
            raise AttributeError("cannot reassign publisher attribute of event {}".format(self.name)) 

    def push(self, data : typing.Any = None, *, serialize : bool = True, **kwargs) -> None:
        """
        publish the event. 

        Parameters
        ----------
        data: Any
            payload of the event
        serialize: bool, default True
            serialize the payload before pushing, set to False when supplying raw bytes
        **kwargs:
            zmq_clients: bool, default True
                pushes event to RPC clients, irrelevant if ``Thing`` uses only one type of serializer (refer to 
                difference between zmq_serializer and http_serializer).
            http_clients: bool, default True
                pushed event to HTTP clients, irrelevant if ``Thing`` uses only one type of serializer (refer to 
                difference between zmq_serializer and http_serializer).
        """
        self.publisher.publish(self._unique_identifier, data, zmq_clients=kwargs.get('zmq_clients', True), 
                                    http_clients=kwargs.get('http_clients', True), serialize=serialize)




class CriticalEvent(Event):
    """
    Push events to client and get acknowledgement for that
    """

    def __init__(self, name : str, URL_path : typing.Optional[str] = None) -> None:
        super().__init__(name, URL_path)
        self._synchronize_event = threading.Event()

    def receive_acknowledgement(self, timeout : typing.Union[float, int, None]) -> bool:
        """
        Receive acknowlegement for event receive. When the timeout argument is present and not None, 
        it should be a floating point number specifying a timeout for the operation in seconds (or fractions thereof).
        """
        return self._synchronize_event.wait(timeout=timeout)

    def _set_acknowledgement(self):
        """
        Method to be called by RPC server when an acknowledgement is received. Not for user to be set.
        """
        self._synchronize_event.set()


__all__ = [
    Event.__name__,
]







def copy_parameters(src : str = 'D:/onedrive/desktop/dashboard/scada/scadapy/scadapy/param/parameters.py', 
                    dst : str = 'D:/onedrive/desktop/dashboard/scada/scadapy/scadapy/server/remote_parameters.py') -> None:

    skip_classes = ['Infinity', 'resolve_path', 'normalize_path', 'BaseConstrainedList', 'TypeConstrainedList', 
                'TypeConstrainedDict', 'TypedKeyMappingsConstrainedDict', 'Event']
    end_line = 'def hashable'
    additional_imports = ['from ..param.parameters import (TypeConstrainedList, TypeConstrainedDict, abbreviate_paths,\n',
                        '                       TypedKeyMappingsConstrainedDict, resolve_path, concrete_descendents, named_objs)\n'
                        'from .remote_parameter import RemoteParameter\n',
                        'from .constants import HTTP, PROXY, USE_OBJECT_NAME, GET, PUT']

    def fetch_line() -> typing.Generator[str]:
        with open(src, 'r') as file:
            oldlines = file.readlines()
            for line in oldlines: 
                yield line

    remote_init_kwargs = [
        '\t\t\tURL_path : str = USE_OBJECT_NAME, http_method : typing.Tuple[str, str] = (GET, PUT),\n', 
        '\t\t\tstate : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,\n',
        '\t\t\tdb_persist : bool = False, db_init : bool = False, db_commit : bool = False,\n'
        '\t\t\taccess_type : str = (HTTP, PROXY),\n',
    ]

    remote_super_init = [
        '\t\t\tURL_path=URL_path, http_method=http_method, state=state, db_persist=db_persist,\n',
        '\t\t\tdb_init=db_init, db_commit=db_commit, access_type=access_type)\n'
    ]

    common_linegen = fetch_line()
    newlines = []

    def skip_to_init_doc():
        for line in common_linegen:
            if 'doc : typing.Optional[str] = None' in line:
                return line
            else:
                newlines.append(line)

    def skip_to_super_init_end():
        for line in common_linegen:
            if 'precedence=precedence)' in line:
                return line
            else:
                newlines.append(line)

    def is_function(line : str) -> bool:
        if 'def ' in line and 'self' not in line and 'cls' not in line and 'obj' not in line:
            return True
        return False 
    
    def next_line_after_skip_class_or_function() -> str:
        for line_ in common_linegen:
            if ('class ' in line_ and ':' in line_) or is_function(line_):
                return line_ 

    def process_current_line(line : str):
        newline = line
        if 'import ' in line and 'parameterized ' in line: 
            newlines_ = [line.replace('from .parameterized', 'from ..param.parameterized').replace('ParamOverrides,', ''), 
                        next(common_linegen).replace('ParameterizedFunction, descendents,', ''), 
                        *additional_imports]
            newlines.extend(newlines_)
            return
        elif 'from collections import OrderedDict' in line:
            newlines.append('from enum import Enum\n')
        elif 'from .utils' in line or 'from .exceptions' in line:
            newline = line.replace('from .', 'from ..param.')
        elif 'class ' in line and ':' in line and line.startswith('class'):
            if '(Parameter):' in line: 
                newline = line.replace('(Parameter):', '(RemoteParameter):') 
                newlines.append(newline)           
            else:
                classname_with_inheritance = line.split(' ', 1)[1][:-2] # [:-2] for removing colon 
                classname_without_inheritance = classname_with_inheritance.split('(', 1)[0]
                if classname_without_inheritance in skip_classes:
                    newline = next_line_after_skip_class_or_function()
                    process_current_line(newline)
                    return
                else:
                    newlines.append(line)
            newline = skip_to_init_doc()
            newlines.append(newline)
            newlines.extend(remote_init_kwargs)
            newline = skip_to_super_init_end()
            if newline:
                newline = newline.replace('precedence=precedence)', 'precedence=precedence,')
                newlines.append(newline)
                newlines.extend(remote_super_init)
            return
        elif 'Parameter.__init__' in line:
            newline = line.replace('Parameter.__init__', 'RemoteParameter.__init__')
        elif is_function(line):
            newline = next_line_after_skip_class_or_function()
            process_current_line(newline)
            return
        newlines.append(newline)


    for line in common_linegen:
        process_current_line(line)
        if end_line in line:
            newlines.pop()
            break
        
    with open(dst, 'w') as file:
        file.writelines(newlines)


"""
The following is a list of all dataclasses used to store information on the exposed 
resources on the network. These classese are generally not for consumption by the package-end-user. 
"""
import typing
import platform
import inspect
from enum import Enum
from dataclasses import dataclass, asdict, field, fields
from types import FunctionType, MethodType

from ..param.parameters import String, Boolean, Tuple, TupleSelector, ClassSelector, Parameter
from ..param.parameterized import ParameterizedMetaclass, ParameterizedFunction
from .constants import JSON, USE_OBJECT_NAME, UNSPECIFIED, HTTP_METHODS, REGEX, ResourceTypes, http_methods 
from .utils import get_signature, getattr_without_descriptor_read, pep8_to_URL_path
from .config import global_config
from .schema_validators import BaseSchemaValidator


class RemoteResourceInfoValidator:
    """
    A validator class for saving remote access related information on a resource. Currently callables (functions, 
    methods and those with__call__) and class/instance property store this information as their own attribute under 
    the variable ``_remote_info``. This is later split into information suitable for HTTP server, RPC client & ``EventLoop``. 
    
    Attributes
    ----------
    URL_path : str, default - extracted object name 
        the path in the URL under which the object is accesible.
        Must follow url-regex ('[\-a-zA-Z0-9@:%._\/\+~#=]{1,256}') requirement. 
        If not specified, the name of object will be used. Underscores will be converted to dashes 
        for PEP 8 names. 
    http_method : str, default POST
        HTTP request method under which the object is accessible. GET, POST, PUT, DELETE or PATCH are supported. 
    state : str, default None
        State machine state at which a callable will be executed or attribute/property can be 
        written. Does not apply to read-only attributes/properties. 
    obj_name : str, default - extracted object name
        the name of the object which will be supplied to the ``ObjectProxy`` class to populate
        its own namespace. For HTTP clients, HTTP method and URL path is important and for 
        object proxies clients, the obj_name is important. 
    iscoroutine : bool, default False 
        whether the callable should be awaited
    isaction : bool, default False 
        True for a method or function or callable
    isproperty : bool, default False
        True for a property
    """
    URL_path = String(default=USE_OBJECT_NAME,
                    doc="the path in the URL under which the object is accesible.") # type: str
    http_method = TupleSelector(default=HTTP_METHODS.POST, objects=http_methods, accept_list=True,
                    doc="HTTP request method under which the object is accessible. GET, POST, PUT, DELETE or PATCH are supported.") # typing.Tuple[str]
    state = Tuple(default=None, item_type=(Enum, str), allow_None=True, accept_list=True, accept_item=True,
                    doc="State machine state at which a callable will be executed or attribute/property can be written.") # type: typing.Union[Enum, str]
    obj = ClassSelector(default=None, allow_None=True, class_=(FunctionType, MethodType, classmethod, Parameter, ParameterizedMetaclass), # Property will need circular import so we stick to base class Parameter
                    doc="the unbound object like the unbound method")
    obj_name = String(default=USE_OBJECT_NAME, 
                    doc="the name of the object which will be supplied to the ``ObjectProxy`` class to populate its own namespace.") # type: str
    isaction = Boolean(default=False,
                    doc="True for a method or function or callable") # type: bool
    isproperty = Boolean(default=False,
                    doc="True for a property") # type: bool
    
    def __init__(self, **kwargs) -> None:
        """   
        No full-scale checks for unknown keyword arguments as the class 
        is used by the developer, so please try to be error-proof
        """
        if kwargs.get('URL_path', None) is not None:
            if not isinstance(kwargs['URL_path'], str): 
                raise TypeError(f"URL path must be a string. Given type {type(kwargs['URL_path'])}")
            if kwargs["URL_path"] != USE_OBJECT_NAME and not kwargs["URL_path"].startswith('/'):
                raise ValueError(f"URL path must start with '/'. Given value {kwargs['URL_path']}")
        for key, value in kwargs.items(): 
            setattr(self, key, value)
    
    def to_dataclass(self, obj : typing.Any = None, bound_obj : typing.Any = None) -> "RemoteResource":
        """
        For a plain, faster and uncomplicated access, a dataclass in created & used by the
        event loop. 
        
        Parameters
        ----------
        obj : Union[Property | Callable]  
            property or method/action

        bound_obj : owner instance
            ``Thing`` instance
       
        Returns
        -------
        RemoteResource
            dataclass equivalent of this object
        """
        return RemoteResource(
                    state=tuple(self.state) if self.state is not None else None, 
                    obj_name=self.obj_name, isaction=self.isaction, 
                    isproperty=self.isproperty, obj=obj, bound_obj=bound_obj, 
                ) 
        # http method is manually always stored as a tuple
    

class ActionInfoValidator(RemoteResourceInfoValidator):
    """
    request_as_argument : bool, default False
        if True, http/RPC request object will be passed as an argument to the callable. 
        The user is warned to not use this generally. 
    argument_schema: JSON, default None
        JSON schema validations for arguments of a callable. Assumption is therefore arguments will be JSON complaint. 
    return_value_schema: JSON, default None 
        schema for return value of a callable. Assumption is therefore return value will be JSON complaint.
    create_task: bool, default True
        default for async methods/actions 
    safe: bool, default True
        metadata information whether the action is safe to execute
    idempotent: bool, default False
        metadata information whether the action is idempotent
    synchronous: bool, default True
        metadata information whether the action is synchronous
    """
    request_as_argument = Boolean(default=False,
                    doc="if True, http/RPC request object will be passed as an argument to the callable.") # type: bool
    argument_schema = ClassSelector(default=None, allow_None=True, class_=dict, 
                    # due to schema validation, this has to be a dict, and not a special dict like TypedDict
                    doc="JSON schema validations for arguments of a callable")
    return_value_schema = ClassSelector(default=None, allow_None=True, class_=dict, 
                    # due to schema validation, this has to be a dict, and not a special dict like TypedDict
                    doc="schema for return value of a callable")
    create_task = Boolean(default=True, 
                        doc="should a coroutine be tasked or run in the same loop?") # type: bool
    iscoroutine = Boolean(default=False, # not sure if isFuture or isCoroutine is correct, something to fix later
                    doc="whether the callable should be awaited") # type: bool
    safe = Boolean(default=True,
                    doc="metadata information whether the action is safe to execute") # type: bool
    idempotent = Boolean(default=False,
                    doc="metadata information whether the action is idempotent") # type: bool
    synchronous = Boolean(default=True,
                    doc="metadata information whether the action is synchronous") # type: bool
    isparameterized = Boolean(default=False,
                    doc="True for a parameterized function") # type: bool


    def to_dataclass(self, obj : typing.Any = None, bound_obj : typing.Any = None) -> "RemoteResource":
        return ActionResource(
                    state=tuple(self.state) if self.state is not None else None, 
                    obj_name=self.obj_name, isaction=self.isaction, iscoroutine=self.iscoroutine,
                    isproperty=self.isproperty, obj=obj, bound_obj=bound_obj, 
                    schema_validator=(bound_obj.schema_validator)(self.argument_schema) if not global_config.validate_schema_on_client and self.argument_schema else None,
                    create_task=self.create_task, isparameterized=self.isparameterized
                ) 
    


class SerializableDataclass:
    """
    Presents uniform serialization for serializers using getstate and setstate and json 
    serialization.
    """
    def json(self):
        return asdict(self)

    def __getstate__(self):
        return self.json()
    
    def __setstate__(self, values : typing.Dict):
        for key, value in values.items():
            setattr(self, key, value)


__dataclass_kwargs = dict(frozen=True)
if float('.'.join(platform.python_version().split('.')[0:2])) >= 3.11:
    __dataclass_kwargs["slots"] = True

@dataclass(**__dataclass_kwargs)
class RemoteResource(SerializableDataclass):
    """
    This container class is used by the ``EventLoop`` methods (for example ``execute_once()``) to access resource 
    metadata instead of directly using ``RemoteResourceInfoValidator``. Instances of this dataclass is stored under 
    ``Thing.instance_resources`` dictionary for each property & method/action. Events use similar dataclass with 
    metadata but with much less information. 

    Attributes
    ----------
    state : str
        State machine state at which a callable will be executed or attribute/property can be 
        written. Does not apply to read-only attributes/properties. 
    obj_name : str, default - extracted object name
        the name of the object which will be supplied to the ``ObjectProxy`` class to populate
        its own namespace. For HTTP clients, HTTP method and URL path is important and for 
        object proxies clients, the obj_name is important. 
    isaction : bool
        True for a method or function or callable
    isproperty : bool
        True for a property
    obj : Union[Property | Callable]  
            property or method/action
    bound_obj : owner instance
        ``Thing`` instance
    """
    state : typing.Optional[typing.Union[typing.Tuple, str]] 
    obj_name : str 
    isaction : bool 
    isproperty : bool
    obj : typing.Any
    bound_obj : typing.Any
    
    def json(self):
        """
        return this object as a JSON serializable dictionary
        """
        # try:
        #     return self._json # accessing dynamic attr from frozen object
        # except AttributeError: # always causes attribute error when slots are True
        json_dict = {}
        for field in fields(self):
            if field.name != 'obj' and field.name != 'bound_obj':
                json_dict[field.name] = getattr(self, field.name)
        # object.__setattr__(self, '_json', json_dict) # because object is frozen - used to work, but not now 
        return json_dict     


@dataclass(**__dataclass_kwargs)
class ActionResource(RemoteResource):  
    """
    Attributes
    ----------
    iscoroutine : bool
        whether the callable should be awaited
    schema_validator : BaseSchemaValidator
        schema validator for the callable if to be validated server side
    """ 
    iscoroutine : bool
    schema_validator : typing.Optional[BaseSchemaValidator]
    create_task : bool 
    isparameterized : bool
    # no need safe, idempotent, synchronous


@dataclass
class HTTPMethodInstructions(SerializableDataclass):
    """
    contains a map of unique strings that identifies the resource operation for each HTTP method, thus acting as 
    instructions to be passed to the RPC server. The unique strings are generally made using the URL_path. 
    """
    GET :  typing.Optional[str] = field(default=None)
    POST :  typing.Optional[str] = field(default=None)
    PUT :  typing.Optional[str] = field(default=None)
    DELETE :  typing.Optional[str] = field(default=None)
    PATCH : typing.Optional[str] = field(default=None) 

    def __post_init__(self):
        self.supported_methods()

    def supported_methods(self): # can be a property
        try: 
            return self._supported_methods
        except: 
            self._supported_methods = []
            for method in ["GET", "POST", "PUT", "DELETE", "PATCH"]:
                if isinstance(self.__dict__[method], str):
                    self._supported_methods.append(method)
            return self._supported_methods
    
    def __contains__(self, value):
        return value in self._supported_methods


@dataclass
class HTTPResource(SerializableDataclass):
    """
    Representation of the resource used by HTTP server for routing and passing information to RPC server on
    "what to do with which resource belonging to which thing? - read, write, execute?". 
    
    Attributes
    ----------

    what : str
        is it a property, method/action or event?
    instance_name : str
        The ``instance_name`` of the thing which owns the resource. Used by HTTP server to inform 
        the message brokers to send the instruction to the correct recipient thing.
    fullpath : str
        URL full path used for routing
    instructions : HTTPMethodInstructions
        unique string that identifies the resource operation for each HTTP method, generally made using the URL_path 
        (qualified URL path {instance name}/{URL path}).  
    argument_schema : JSON
        argument schema of the method/action for validation before passing over the instruction to the RPC server. 
    request_as_argument: bool
        pass the request as a argument to the callable. For HTTP server ``tornado.web.HTTPServerRequest`` will be passed. 
    """
    what : str 
    class_name : str # just metadata
    instance_name : str 
    obj_name : str
    fullpath : str
    instructions : HTTPMethodInstructions
    argument_schema : typing.Optional[JSON]
    request_as_argument : bool = field(default=False)
    
                                  
    def __init__(self, *, what : str, class_name : str, instance_name : str, obj_name : str, fullpath : str, 
                request_as_argument : bool = False, argument_schema : typing.Optional[JSON] = None, 
                **instructions) -> None:
        self.what = what 
        self.class_name = class_name
        self.instance_name = instance_name
        self.obj_name = obj_name
        self.fullpath = fullpath
        self.request_as_argument = request_as_argument
        self.argument_schema = argument_schema
        if instructions.get('instructions', None):
            self.instructions = HTTPMethodInstructions(**instructions.get('instructions', None))
        else: 
            self.instructions = HTTPMethodInstructions(**instructions)
    

@dataclass
class ZMQResource(SerializableDataclass): 
    """
    Representation of resource used by RPC clients for mapping client method/action calls, property read/writes & events
    to a server resource. Used to dynamically populate the ``ObjectProxy``

    Attributes
    ----------

    what : str
        is it a property, method/action or event?
    instance_name : str
        The ``instance_name`` of the thing which owns the resource. Used by RPC client to inform 
        message brokers to send the message to the correct recipient.
    instruction : str
        unique string that identifies the resource, generally made using the URL_path. Although URL path is a HTTP
        concept, it is still used as a unique identifier. 
    name : str
        the name of the resource (__name__)
    qualname : str
        the qualified name of the resource (__qualname__) 
    doc : str
        the docstring of the resource
    argument_schema : JSON
        argument schema of the method/action for validation before passing over the instruction to the RPC server. 
    """
    what : str 
    class_name : str # just metadata
    instance_name : str 
    instruction : str
    obj_name : str
    qualname : str
    doc : typing.Optional[str]
    top_owner : bool 
    argument_schema : typing.Optional[JSON]
    return_value_schema : typing.Optional[JSON]
    request_as_argument : bool = field(default=False)

    def __init__(self, *, what : str, class_name : str, instance_name : str, instruction : str, obj_name : str,
                qualname : str, doc : str, top_owner : bool, argument_schema : typing.Optional[JSON] = None,
                return_value_schema : typing.Optional[JSON] = None, request_as_argument : bool = False) -> None:
        self.what = what 
        self.class_name = class_name
        self.instance_name = instance_name
        self.instruction = instruction
        self.obj_name = obj_name 
        self.qualname = qualname
        self.doc = doc
        self.top_owner = top_owner
        self.argument_schema = argument_schema
        self.return_value_schema = return_value_schema
        self.request_as_argument = request_as_argument

    def get_dunder_attr(self, __dunder_name : str):
        name = __dunder_name.strip('_')
        name = 'obj_name' if name == 'name' else name
        return getattr(self, name)


@dataclass
class ServerSentEvent(SerializableDataclass):
    """
    event name and socket address of events to be consumed by clients. 
  
    Attributes
    ----------
    name : str
        name of the event, must be unique
    obj_name: str
        name of the event variable used to populate the RPC client
    socket_address : str
        address of the socket
    unique_identifier: str
        unique ZMQ identifier used in PUB-SUB model
    what: str, default EVENT
        is it a property, method/action or event?
    """
    name : str = field(default=UNSPECIFIED)
    obj_name : str = field(default=UNSPECIFIED)
    class_name : str = field(default=UNSPECIFIED) # just metadata  
    unique_identifier : str = field(default=UNSPECIFIED)
    serialization_specific : bool = field(default=False)
    socket_address : str = field(default=UNSPECIFIED)
    what : str = field(default=ResourceTypes.EVENT)


def build_our_temp_TD(instance):
    """
    A temporary extension of TD used to build GUI of thing control panel.
    Will be later replaced by a more sophisticated TD builder which is compliant to the actual spec & its theory.
    """
    from .thing import Thing

    assert isinstance(instance, Thing), f"got invalid type {type(instance)}"
    
    our_TD = instance.get_thing_description(ignore_errors=True)
    our_TD["inheritance"] = [class_.__name__ for class_ in instance.__class__.mro()]

    for instruction, remote_info in instance.instance_resources.items(): 
        if remote_info.isaction and remote_info.obj_name in our_TD["actions"]:
            if isinstance(remote_info.obj, classmethod):
                our_TD["actions"][remote_info.obj_name]["type"] = 'classmethod'
            our_TD["actions"][remote_info.obj_name]["signature"] = get_signature(remote_info.obj)[0]
        elif remote_info.isproperty and remote_info.obj_name in our_TD["properties"]:
            our_TD["properties"][remote_info.obj_name].update(instance.__class__.properties.webgui_info(remote_info.obj)[remote_info.obj_name])
    return our_TD



def get_organised_resources(instance):
    """
    organise the exposed attributes, actions and events into the dataclasses defined above
    so that the specific servers and event loop can use them. 
    """
    from .thing import Thing
    from .events import Event, EventDispatcher
    from .property import Property

    assert isinstance(instance, Thing), f"got invalid type {type(instance)}"

    httpserver_resources = dict() # type: typing.Dict[str, HTTPResource]
    # The following dict will be given to the object proxy client
    zmq_resources = dict() # type: typing.Dict[str, ZMQResource]
    # The following dict will be used by the event loop
    instance_resources = dict() # type: typing.Dict[str, typing.Union[RemoteResource, ActionResource]] 
    # create URL prefix
    if instance._owner is not None: 
        instance._full_URL_path_prefix = f'{instance._owner._full_URL_path_prefix}/{instance.instance_name}' 
    else:
        instance._full_URL_path_prefix = f'/{instance.instance_name}' # leading '/' was stripped at init
    
    # First add methods and callables
    # properties
    for prop in instance.parameters.descriptors.values():
        if isinstance(prop, Property) and hasattr(prop, '_remote_info') and prop._remote_info is not None: 
            if not isinstance(prop._remote_info, RemoteResourceInfoValidator): 
                raise TypeError("instance member {} has unknown sub-member '_remote_info' of type {}.".format(
                            prop, type(prop._remote_info))) 
                # above condition is just a gaurd in case somebody does some unpredictable patching activities
            remote_info = prop._remote_info
            fullpath = f"{instance._full_URL_path_prefix}{remote_info.URL_path}"
            read_http_method = write_http_method = delete_http_method = None
            if len(remote_info.http_method) == 1:
                read_http_method = remote_info.http_method[0]
                instructions = { read_http_method : f"{fullpath}/read" }
            elif len(remote_info.http_method) == 2:
                read_http_method, write_http_method = remote_info.http_method
                instructions = { 
                    read_http_method : f"{fullpath}/read", 
                    write_http_method : f"{fullpath}/write"
                }
            else:
                read_http_method, write_http_method, delete_http_method = remote_info.http_method
                instructions = {
                    read_http_method : f"{fullpath}/read", 
                    write_http_method : f"{fullpath}/write", 
                    delete_http_method : f"{fullpath}/delete"
                }
                
            httpserver_resources[fullpath] = HTTPResource(
                                                what=ResourceTypes.PROPERTY, 
                                                class_name=instance.__class__.__name__,
                                                instance_name=instance._owner.instance_name if instance._owner is not None else instance.instance_name,
                                                obj_name=remote_info.obj_name,
                                                fullpath=fullpath,
                                                **instructions
                                            )
            zmq_resources[fullpath] = ZMQResource(
                            what=ResourceTypes.PROPERTY, 
                            class_name=instance.__class__.__name__,
                            instance_name=instance._owner.instance_name if instance._owner is not None else instance.instance_name, 
                            instruction=fullpath, 
                            doc=prop.__doc__, 
                            obj_name=remote_info.obj_name,
                            qualname=instance.__class__.__name__ + '.' + remote_info.obj_name,
                            # qualname is not correct probably, does not respect inheritance
                            top_owner=instance._owner is None,
                        ) 
            data_cls = remote_info.to_dataclass(obj=prop, bound_obj=instance) 
            instance_resources[f"{fullpath}/read"] = data_cls
            instance_resources[f"{fullpath}/write"] = data_cls  
            instance_resources[f"{fullpath}/delete"] = data_cls  
            if prop._observable:
                # There is no real philosophy behind this logic flow, we just set the missing information.
                assert isinstance(prop._observable_event_descriptor, Event), f"observable event not yet set for {prop.name}. logic error."
                evt_fullpath = f"{instance._full_URL_path_prefix}{prop._observable_event_descriptor.URL_path}"
                dispatcher = EventDispatcher(evt_fullpath)
                dispatcher._remote_info.class_name = instance.__class__.__name__
                dispatcher._remote_info.serialization_specific = instance.zmq_serializer != instance.http_serializer
                setattr(instance, prop._observable_event_descriptor._obj_name, dispatcher)
                # prop._observable_event_descriptor._remote_info.unique_identifier = evt_fullpath
                httpserver_resources[evt_fullpath] = dispatcher._remote_info
                zmq_resources[evt_fullpath] = dispatcher._remote_info
    # Methods
    for name, resource in inspect._getmembers(instance, lambda f : inspect.ismethod(f) or (
                                hasattr(f, '_remote_info') and isinstance(f._remote_info, ActionInfoValidator)),
                                                 getattr_without_descriptor_read): 
        if hasattr(resource, '_remote_info'):
            if not isinstance(resource._remote_info, ActionInfoValidator):
                raise TypeError("instance member {} has unknown sub-member '_remote_info' of type {}.".format(
                            resource, type(resource._remote_info))) 
            remote_info = resource._remote_info
            # methods are already bound
            assert remote_info.isaction, ("remote info from inspect.ismethod is not a callable",
                                "logic error - visit https://github.com/VigneshVSV/hololinked/issues to report")
            fullpath = f"{instance._full_URL_path_prefix}{remote_info.URL_path}" 
            instruction = f"{fullpath}/invoke-on-{remote_info.http_method[0]}" 
            # needs to be cleaned up for multiple HTTP methods
            httpserver_resources[instruction] = HTTPResource(
                                        what=ResourceTypes.ACTION,
                                        class_name=instance.__class__.__name__,
                                        instance_name=instance._owner.instance_name if instance._owner is not None else instance.instance_name,
                                        obj_name=remote_info.obj_name,
                                        fullpath=fullpath,
                                        argument_schema=remote_info.argument_schema,
                                        request_as_argument=remote_info.request_as_argument,
                                        **{ http_method : instruction for http_method in remote_info.http_method },
                                    )
            zmq_resources[instruction] = ZMQResource(
                                            what=ResourceTypes.ACTION,
                                            class_name=instance.__class__.__name__,
                                            instance_name=instance._owner.instance_name if instance._owner is not None else instance.instance_name,
                                            instruction=instruction,                                                                                                                                                                                        
                                            obj_name=getattr(resource, '__name__'),
                                            qualname=getattr(resource, '__qualname__'), 
                                            doc=getattr(resource, '__doc__'),
                                            top_owner=instance._owner is None,
                                            argument_schema=remote_info.argument_schema,
                                            return_value_schema=remote_info.return_value_schema,
                                            request_as_argument=remote_info.request_as_argument
                                        )
            instance_resources[instruction] = remote_info.to_dataclass(obj=resource, bound_obj=instance)   
    # Events
    for name, resource in inspect._getmembers(instance, lambda o : isinstance(o, Event), getattr_without_descriptor_read):
        assert isinstance(resource, Event), ("thing event query from inspect.ismethod is not an Event",
                                    "logic error - visit https://github.com/VigneshVSV/hololinked/issues to report")    
        # above assertion is only a typing convenience
        fullpath = f"{instance._full_URL_path_prefix}{resource.URL_path}"
        # resource._remote_info.unique_identifier = fullpath
        dispatcher = EventDispatcher(fullpath)
        dispatcher._remote_info.class_name = instance.__class__.__name__
        dispatcher._remote_info.serialization_specific = instance.zmq_serializer != instance.http_serializer
        setattr(instance, name, dispatcher) # resource._remote_info.unique_identifier))
        httpserver_resources[fullpath] = dispatcher._remote_info
        zmq_resources[fullpath] = dispatcher._remote_info
    # Other objects
    for name, resource in inspect._getmembers(instance, lambda o : isinstance(o, Thing), getattr_without_descriptor_read):
        assert isinstance(resource, Thing), ("thing children query from inspect.ismethod is not a Thing",
                                    "logic error - visit https://github.com/VigneshVSV/hololinked/issues to report")
        # above assertion is only a typing convenience
        if name == '_owner' or resource._owner is not None: 
            # second condition allows sharing of Things without adding once again to the list of exposed resources
            # for example, a shared logger 
            continue
        resource._owner = instance      
        resource._prepare_resources()
        httpserver_resources.update(resource.httpserver_resources)
        # zmq_resources.update(resource.zmq_resources)
        instance_resources.update(resource.instance_resources)
   
    # The above for-loops can be used only once, the division is only for readability
    # following are in _internal_fixed_attributes - allowed to set only once
    return zmq_resources, httpserver_resources, instance_resources    


import threading 
import warnings
import typing 
import logging
import uuid


from ..server.config import global_config
from ..server.constants import JSON, CommonRPC, ServerMessage, ResourceTypes, ZMQ_PROTOCOLS
from ..server.serializers import BaseSerializer
from ..server.dataklasses import ZMQResource, ServerSentEvent
from ..server.zmq_message_brokers import AsyncZMQClient, SyncZMQClient, EventConsumer, PROXY
from ..server.schema_validators import BaseSchemaValidator


class ObjectProxy:
    """
    Procedural client for ``Thing``/``RemoteObject``. Once connected to a server, properties, methods and events are 
    dynamically populated. Any of the ZMQ protocols of the server is supported. 

    Parameters
    ----------
    instance_name: str
        instance name of the server to connect.
    invokation_timeout: float, int
        timeout to schedule a method call or property read/write in server. execution time wait is controlled by 
        ``execution_timeout``. When invokation timeout expires, the method is not executed. 
    execution_timeout: float, int
        timeout to return without a reply after scheduling a method call or property read/write. This timer starts
        ticking only after the method has started to execute. Returning a call before end of execution can lead to 
        change of state in the server. 
    load_thing: bool, default True
        when True, remote object is located and its resources are loaded. Otherwise, only the client is initialised.
    protocol: str
        ZMQ protocol used to connect to server. Unlike the server, only one can be specified.  
    **kwargs: 
        async_mixin: bool, default False
            whether to use both synchronous and asynchronous clients. 
        serializer: BaseSerializer
            use a custom serializer, must be same as the serializer supplied to the server. 
        schema_validator: BaseSchemaValidator
            use a schema validator, must be same as the schema validator supplied to the server.
        allow_foreign_attributes: bool, default False
            allows local attributes for proxy apart from properties fetched from the server.
        logger: logging.Logger
            logger instance
        log_level: int
            log level corresponding to logging.Logger when internally created
        handshake_timeout: int
            time in milliseconds to search & handshake server remote object. raises Timeout when expired
    """

    _own_attrs = frozenset([
        '__annotations__',
        'zmq_client', 'async_zmq_client', '_allow_foreign_attributes',
        'identity', 'instance_name', 'logger', 'execution_timeout', 'invokation_timeout', 
        '_execution_timeout', '_invokation_timeout', '_events', '_noblock_messages',
        '_schema_validator'
    ])

    def __init__(self, instance_name : str, protocol : str = ZMQ_PROTOCOLS.IPC, invokation_timeout : float = 5, 
                    load_thing = True, **kwargs) -> None:
        self._allow_foreign_attributes = kwargs.get('allow_foreign_attributes', False)
        self.instance_name = instance_name
        self.invokation_timeout = invokation_timeout
        self.execution_timeout = kwargs.get("execution_timeout", None)
        self.identity = f"{instance_name}|{uuid.uuid4()}"
        self.logger = kwargs.pop('logger', logging.Logger(self.identity, level=kwargs.get('log_level', logging.INFO)))
        self._noblock_messages = dict()
        self._schema_validator = kwargs.get('schema_validator', None)
        # compose ZMQ client in Proxy client so that all sending and receiving is
        # done by the ZMQ client and not by the Proxy client directly. Proxy client only 
        # bothers mainly about __setattr__ and _getattr__
        self.async_zmq_client = None    
        self.zmq_client = SyncZMQClient(instance_name, self.identity, client_type=PROXY, protocol=protocol, 
                                            zmq_serializer=kwargs.get('serializer', None), handshake=load_thing,
                                            logger=self.logger, **kwargs)
        if kwargs.get("async_mixin", False):
            self.async_zmq_client = AsyncZMQClient(instance_name, self.identity + '|async', client_type=PROXY, protocol=protocol, 
                                            zmq_serializer=kwargs.get('serializer', None), handshake=load_thing,
                                            logger=self.logger, **kwargs)
        if load_thing:
            self.load_thing()

    def __getattribute__(self, __name: str) -> typing.Any:
        obj = super().__getattribute__(__name)
        if isinstance(obj, _Property):
            return obj.get()
        return obj

    def __setattr__(self, __name : str, __value : typing.Any) -> None:
        if (__name in ObjectProxy._own_attrs or (__name not in self.__dict__ and 
                isinstance(__value, __allowed_attribute_types__)) or self._allow_foreign_attributes):
            # allowed attribute types are _Property and _RemoteMethod defined after this class
            return super(ObjectProxy, self).__setattr__(__name, __value)
        elif __name in self.__dict__:
            obj = self.__dict__[__name]
            if isinstance(obj, _Property):
                obj.set(value=__value)
                return
            raise AttributeError(f"Cannot set attribute {__name} again to ObjectProxy for {self.instance_name}.")
        raise AttributeError(f"Cannot set foreign attribute {__name} to ObjectProxy for {self.instance_name}. Given attribute not found in server object.")

    def __repr__(self) -> str:
        return f'ObjectProxy {self.identity}'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __bool__(self) -> bool: 
        try: 
            self.zmq_client.handshake(num_of_tries=10)
            return True
        except RuntimeError:
            return False

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        return (isinstance(other, ObjectProxy) and other.instance_name == self.instance_name and 
                other.zmq_client.protocol == self.zmq_client.protocol)

    def __ne__(self, other) -> bool:
        if other and isinstance(other, ObjectProxy):
            return (other.instance_name != self.instance_name or 
                    other.zmq_client.protocol != self.zmq_client.protocol)
        return True

    def __hash__(self) -> int:
        return hash(self.identity)
    
    def get_invokation_timeout(self) -> typing.Union[float, int]:
        return self._invokation_timeout 
    
    def set_invokation_timeout(self, value : typing.Union[float, int]) -> None:
        if not isinstance(value, (float, int, type(None))):
            raise TypeError(f"Timeout can only be float or int greater than 0, or None. Given type {type(value)}.")
        elif value is not None and value < 0:
            raise ValueError("Timeout must be at least 0 or None, not negative.")
        self._invokation_timeout = value
    
    invokation_timeout = property(fget=get_invokation_timeout, fset=set_invokation_timeout,
                        doc="Timeout in seconds on server side for invoking a method or read/write property. \
                                Defaults to 5 seconds and network times not considered."
    )

    def get_execution_timeout(self) -> typing.Union[float, int]:
        return self._execution_timeout 
    
    def set_execution_timeout(self, value : typing.Union[float, int]) -> None:
        if not isinstance(value, (float, int, type(None))):
            raise TypeError(f"Timeout can only be float or int greater than 0, or None. Given type {type(value)}.")
        elif value is not None and value < 0:
            raise ValueError("Timeout must be at least 0 or None, not negative.")
        self._execution_timeout = value
    
    execution_timeout = property(fget=get_execution_timeout, fset=set_execution_timeout,
                            doc="Timeout in seconds on server side for execution of method or read/write property." +
                                "Starts ticking after invokation timeout completes." + 
                                "Defaults to None (i.e. waits indefinitely until return) and network times not considered."
    )


    def invoke_action(self, method : str, oneway : bool = False, noblock : bool = False, 
                                *args, **kwargs) -> typing.Any:
        """
        call a method specified by name on the server with positional/keyword arguments

        Parameters
        ----------
        method: str 
            name of the method
        oneway: bool, default False 
            only send an instruction to invoke the method but do not fetch the reply.
        noblock: bool, default False 
            request a method call but collect the reply later using a reply id
        *args: Any
            arguments for the method 
        **kwargs: Dict[str, Any]
            keyword arguments for the method

        Returns
        -------
        Any 
            return value of the method call or an id if noblock is True 

        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated 
        """
        method = getattr(self, method, None) # type: _RemoteMethod 
        if not isinstance(method, _RemoteMethod):
            raise AttributeError(f"No remote method named {method}")
        if oneway:
            method.oneway(*args, **kwargs)
        elif noblock:
            msg_id = method.noblock(*args, **kwargs)
            self._noblock_messages[msg_id] = method
            return msg_id
        else:
            return method(*args, **kwargs)


    async def async_invoke_action(self, method : str, *args, **kwargs) -> typing.Any:
        """
        async(io) call a method specified by name on the server with positional/keyword 
        arguments. noblock and oneway not supported for async calls. 

        Parameters
        ----------
        method: str 
            name of the method
        *args: Any
            arguments for the method 
        **kwargs: Dict[str, Any]
            keyword arguments for the method

        Returns
        -------
        Any 
            return value of the method call
        
        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        RuntimeError:
            if async_mixin was False at ``__init__()`` - no asynchronous client was created
        Exception:
            server raised exception are propagated
        """
        method = getattr(self, method, None) # type: _RemoteMethod 
        if not isinstance(method, _RemoteMethod):
            raise AttributeError(f"No remote method named {method}")
        return await method.async_call(*args, **kwargs)


    def read_property(self, name : str, noblock : bool = False) -> typing.Any:
        """
        get property specified by name on server. 

        Parameters
        ----------
        name: str 
            name of the property
        noblock: bool, default False 
            request the property get but collect the reply/value later using a reply id

        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None) # type: _Property
        if not isinstance(prop, _Property):
            raise AttributeError(f"No property named {prop}")
        if noblock:
            msg_id = prop.noblock_get()
            self._noblock_messages[msg_id] = prop
            return msg_id
        else:
            return prop.get()


    def write_property(self, name : str, value : typing.Any, oneway : bool = False, 
                        noblock : bool = False) -> None:
        """
        set property specified by name on server with specified value. 

        Parameters
        ----------
        name: str
            name of the property
        value: Any 
            value of property to be set
        oneway: bool, default False 
            only send an instruction to set the property but do not fetch the reply.
            (irrespective of whether set was successful or not)
        noblock: bool, default False 
            request the set property but collect the reply later using a reply id

        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None) # type: _Property
        if not isinstance(prop, _Property):
            raise AttributeError(f"No property named {prop}")
        if oneway:
            prop.oneway_set(value)
        elif noblock:
            msg_id = prop.noblock_set(value)
            self._noblock_messages[msg_id] = prop
            return msg_id
        else:
            prop.set(value)


    async def async_read_property(self, name : str) -> None:
        """
        async(io) get property specified by name on server. 

        Parameters
        ----------
        name: Any 
            name of the property to fetch 

        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None) # type: _Property
        if not isinstance(prop, _Property):
            raise AttributeError(f"No property named {prop}")
        return await prop.async_get()
    

    async def async_write_property(self, name : str, value : typing.Any) -> None:
        """
        async(io) set property specified by name on server with specified value.  
        noblock and oneway not supported for async calls. 

        Parameters
        ----------
        name: str 
            name of the property
        value: Any 
            value of property to be set
        
        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None) # type: _Property
        if not isinstance(prop, _Property):
            raise AttributeError(f"No property named {prop}")
        await prop.async_set(value)


    def read_multiple_properties(self, names : typing.List[str], noblock : bool = False) -> typing.Any:
        """
        get properties specified by list of names.

        Parameters
        ----------
        names: List[str]
            names of properties to be fetched 
        noblock: bool, default False 
            request the fetch but collect the reply later using a reply id

        Returns
        -------
        Dict[str, Any]:
            dictionary with names as keys and values corresponding to those keys
        """
        method = getattr(self, '_get_properties', None) # type: _RemoteMethod
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        if noblock:
            msg_id = method.noblock(names=names)
            self._noblock_messages[msg_id] = method
            return msg_id
        else:
            return method(names=names)
        
    
    def write_multiple_properties(self, oneway : bool = False, noblock : bool = False,
                       **properties : typing.Dict[str, typing.Any]) -> None:
        """
        set properties whose name is specified by keys of a dictionary

        Parameters
        ----------
        oneway: bool, default False 
            only send an instruction to set the property but do not fetch the reply.
            (irrespective of whether set was successful or not)
        noblock: bool, default False 
            request the set property but collect the reply later using a reply id
        **properties: Dict[str, Any]
            name and value of properties to be set

        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        if len(properties) == 0:
            raise ValueError("no properties given to set_properties")
        method = getattr(self, '_set_properties', None) # type: _RemoteMethod
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        if oneway:
            method.oneway(**properties)
        elif noblock:
            msg_id = method.noblock(**properties)
            self._noblock_messages[msg_id] = method
            return msg_id
        else:
            return method(**properties)
        

    async def async_read_multiple_properties(self, names) -> None:
        """
        async(io) get properties specified by list of names. no block gets are not supported for asyncio.

        Parameters
        ----------
        names: List[str]
            names of properties to be fetched 
 
        Returns
        -------
        Dict[str, Any]:
            dictionary with property names as keys and values corresponding to those keys
        """
        method = getattr(self, '_get_properties', None) # type: _RemoteMethod
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        return await method.async_call(names=names)


    async def async_write_multiple_properties(self, **properties) -> None:
        """
        async(io) set properties whose name is specified by keys of a dictionary

        Parameters
        ----------
        values: Dict[str, Any] 
            name and value of properties to be set
       
        Raises
        ------
        AttributeError: 
            if no method with specified name found on the server
        Exception:
            server raised exception are propagated
        """
        if len(properties) == 0:
            raise ValueError("no properties given to set_properties")
        method = getattr(self, '_set_properties', None) # type: _RemoteMethod
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        await method.async_call(**properties)


    def subscribe_event(self, name : str, callbacks : typing.Union[typing.List[typing.Callable], typing.Callable],
                        thread_callbacks : bool = False, deserialize : bool = True) -> None:
        """
        Subscribe to event specified by name. Events are listened in separate threads and supplied callbacks are
        are also called in those threads. 

        Parameters
        ----------
        name: str
            name of the event, either the object name used in the server or the name specified in the name argument of
            the Event object 
        callbacks: Callable | List[Callable]
            one or more callbacks that will be executed when this event is received
        thread_callbacks: bool
            thread the callbacks otherwise the callbacks will be executed serially
        
        Raises
        ------
        AttributeError: 
            if no event with specified name is found
        """
        event = getattr(self, name, None) # type: _Event
        if not isinstance(event, _Event):
            raise AttributeError(f"No event named {name}")
        if event._subscribed:
            event.add_callbacks(callbacks)
        else: 
            event.subscribe(callbacks, thread_callbacks, deserialize)
       

    def unsubscribe_event(self, name : str):
        """
        Unsubscribe to event specified by name. 

        Parameters
        ----------
        name: str
            name of the event 
        callbacks: Callable | List[Callable]
            one or more callbacks that will be executed when this event is received
        thread_callbacks: bool
            thread the callbacks otherwise the callbacks will be executed serially
        
        Raises
        ------
        AttributeError: 
            if no event with specified name is found
        """
        event = getattr(self, name, None) # type: _Event
        if not isinstance(event, _Event):
            raise AttributeError(f"No event named {name}")
        event.unsubscribe()

    
    def read_reply(self, message_id : bytes, timeout : typing.Optional[float] = 5000) -> typing.Any:
        """
        read reply of no block calls of an action or a property read/write.
        """
        obj = self._noblock_messages.get(message_id, None) 
        if not obj:
            raise ValueError('given message id not a one way call or invalid.')
        reply = self.zmq_client._reply_cache.get(message_id, None)
        if not reply: 
            reply = self.zmq_client.recv_reply(message_id=message_id, timeout=timeout,
                                    raise_client_side_exception=True)
        if not reply:
            raise ReplyNotArrivedError(f"could not fetch reply within timeout for message id '{message_id}'")
        if isinstance(obj, _RemoteMethod):
            obj._last_return_value = reply 
            return obj.last_return_value # note the missing underscore
        elif isinstance(obj, _Property):
            obj._last_value = reply 
            return obj.last_read_value


    def load_thing(self):
        """
        Get exposed resources from server (methods, properties, events) and remember them as attributes of the proxy.
        """
        fetch = _RemoteMethod(self.zmq_client, CommonRPC.zmq_resource_read(instance_name=self.instance_name), 
                                    invokation_timeout=self._invokation_timeout) # type: _RemoteMethod
        reply = fetch() # type: typing.Dict[str, typing.Dict[str, typing.Any]]

        for name, data in reply.items():
            if isinstance(data, dict):
                try:
                    if data["what"] == ResourceTypes.EVENT:
                        data = ServerSentEvent(**data)
                    else:
                        data = ZMQResource(**data)
                except Exception as ex:
                    ex.add_note("Did you correctly configure your serializer? " + 
                            "This exception occurs when given serializer does not work the same way as server serializer")
                    raise ex from None
            elif not isinstance(data, (ZMQResource, ServerSentEvent)):
                raise RuntimeError("Logic error - deserialized info about server not instance of hololinked.server.data_classes.ZMQResource")
            if data.what == ResourceTypes.ACTION:
                _add_method(self, _RemoteMethod(self.zmq_client, data.instruction, self.invokation_timeout, 
                                                self.execution_timeout, data.argument_schema, self.async_zmq_client, self._schema_validator), data)
            elif data.what == ResourceTypes.PROPERTY:
                _add_property(self, _Property(self.zmq_client, data.instruction, self.invokation_timeout,
                                                self.execution_timeout, self.async_zmq_client), data)
            elif data.what == ResourceTypes.EVENT:
                assert isinstance(data, ServerSentEvent)
                event = _Event(self.zmq_client, data.name, data.obj_name, data.unique_identifier, data.socket_address, 
                            serialization_specific=data.serialization_specific, serializer=self.zmq_client.zmq_serializer, logger=self.logger)
                _add_event(self, event, data)
                self.__dict__[data.name] = event 


    


# SM = Server Message
SM_INDEX_ADDRESS = ServerMessage.ADDRESS.value
SM_INDEX_SERVER_TYPE = ServerMessage.SERVER_TYPE.value
SM_INDEX_MESSAGE_TYPE = ServerMessage.MESSAGE_TYPE.value
SM_INDEX_MESSAGE_ID = ServerMessage.MESSAGE_ID.value
SM_INDEX_DATA = ServerMessage.DATA.value
SM_INDEX_ENCODED_DATA = ServerMessage.ENCODED_DATA.value

class _RemoteMethod:
    
    __slots__ = ['_zmq_client', '_async_zmq_client', '_instruction', '_invokation_timeout', '_execution_timeout',
                 '_schema', '_schema_validator', '_last_return_value', '__name__', '__qualname__', '__doc__']
    # method call abstraction
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    def __init__(self, sync_client : SyncZMQClient, instruction : str, invokation_timeout : typing.Optional[float] = 5, 
                    execution_timeout : typing.Optional[float] = None, argument_schema : typing.Optional[JSON] = None,
                    async_client : typing.Optional[AsyncZMQClient] = None, 
                    schema_validator : typing.Optional[typing.Type[BaseSchemaValidator]] = None) -> None:
        """
        Parameters
        ----------
        sync_client: SyncZMQClient
            synchronous ZMQ client
        async_zmq_client: AsyncZMQClient
            asynchronous ZMQ client for async calls
        instruction: str
            The instruction needed to call the method        
        """
        self._zmq_client = sync_client
        self._async_zmq_client = async_client
        self._instruction = instruction
        self._invokation_timeout = invokation_timeout
        self._execution_timeout = execution_timeout
        self._schema = argument_schema
        self._schema_validator = schema_validator(self._schema) if schema_validator and argument_schema and global_config.validate_schema_on_client else None
    
    @property # i.e. cannot have setter
    def last_return_value(self):
        """
        cached return value of the last call to the method
        """
        if len(self._last_return_value[SM_INDEX_ENCODED_DATA]) > 0:
            return self._last_return_value[SM_INDEX_ENCODED_DATA]
        return self._last_return_value[SM_INDEX_DATA]
    
    @property
    def last_zmq_message(self) -> typing.List:
        return self._last_return_value
    
    def __call__(self, *args, **kwargs) -> typing.Any:
        """
        execute method on server
        """
        if len(args) > 0: 
            kwargs["__args__"] = args
        elif self._schema_validator:
            self._schema_validator.validate(kwargs)
        self._last_return_value = self._zmq_client.execute(instruction=self._instruction, arguments=kwargs, 
                                    invokation_timeout=self._invokation_timeout, execution_timeout=self._execution_timeout,
                                    raise_client_side_exception=True, argument_schema=self._schema)
        return self.last_return_value # note the missing underscore
    
    def oneway(self, *args, **kwargs) -> None:
        """
        only issues the method call to the server and does not wait for reply,
        neither does the server reply to this call.  
        """
        if len(args) > 0: 
            kwargs["__args__"] = args
        elif self._schema_validator:
            self._schema_validator.validate(kwargs)
        self._zmq_client.send_instruction(instruction=self._instruction, arguments=kwargs, 
                                        invokation_timeout=self._invokation_timeout, execution_timeout=None,
                                        context=dict(oneway=True), argument_schema=self._schema)

    def noblock(self, *args, **kwargs) -> None:
        if len(args) > 0: 
            kwargs["__args__"] = args
        elif self._schema_validator:
            self._schema_validator.validate(kwargs)
        return self._zmq_client.send_instruction(instruction=self._instruction, arguments=kwargs, 
                                invokation_timeout=self._invokation_timeout, execution_timeout=self._execution_timeout,
                                argument_schema=self._schema)
     
    async def async_call(self, *args, **kwargs):
        """
        async execute method on server
        """
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set at __init__()")
        if len(args) > 0: 
            kwargs["__args__"] = args
        elif self._schema_validator:
            self._schema_validator.validate(kwargs)
        self._last_return_value = await self._async_zmq_client.async_execute(instruction=self._instruction, 
                                        arguments=kwargs, invokation_timeout=self._invokation_timeout, 
                                        raise_client_side_exception=True,
                                        argument_schema=self._schema)
        return self.last_return_value # note the missing underscore

    
class _Property:

    __slots__ = ['_zmq_client', '_async_zmq_client', '_read_instruction', '_write_instruction', 
                '_invokation_timeout', '_execution_timeout', '_last_value', '__name__', '__doc__']   
    # property get set abstraction
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    def __init__(self, client : SyncZMQClient, instruction : str, invokation_timeout : typing.Optional[float] = 5, 
                    execution_timeout : typing.Optional[float] = None, async_client : typing.Optional[AsyncZMQClient] = None) -> None:
        self._zmq_client = client
        self._async_zmq_client = async_client
        self._invokation_timeout = invokation_timeout
        self._execution_timeout = execution_timeout
        self._read_instruction = instruction + '/read'
        self._write_instruction = instruction + '/write'

    @property # i.e. cannot have setter
    def last_read_value(self) -> typing.Any:
        """
        cache of last read value
        """
        if len(self._last_value[SM_INDEX_ENCODED_DATA]) > 0:
            return self._last_value[SM_INDEX_ENCODED_DATA]
        return self._last_value[SM_INDEX_DATA]
    
    @property
    def last_zmq_message(self) -> typing.List:
        """
        cache of last message received for this property
        """
        return self._last_value
    
    def set(self, value : typing.Any) -> None:
        self._last_value = self._zmq_client.execute(self._write_instruction, dict(value=value),
                                                        raise_client_side_exception=True)
     
    def get(self) -> typing.Any:
        self._last_value = self._zmq_client.execute(self._read_instruction, 
                                                invokation_timeout=self._invokation_timeout, 
                                                raise_client_side_exception=True)
        return self.last_read_value 
    
    async def async_set(self, value : typing.Any) -> None:
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set at __init__()")
        self._last_value = await self._async_zmq_client.async_execute(self._write_instruction, dict(value=value),
                                                        invokation_timeout=self._invokation_timeout, 
                                                        execution_timeout=self._execution_timeout,
                                                        raise_client_side_exception=True)
    
    async def async_get(self) -> typing.Any:
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set at __init__()")
        self._last_value = await self._async_zmq_client.async_execute(self._read_instruction,
                                                invokation_timeout=self._invokation_timeout, 
                                                execution_timeout=self._execution_timeout,
                                                raise_client_side_exception=True)
        return self.last_read_value 
    
    def noblock_get(self) -> None:
        return self._zmq_client.send_instruction(self._read_instruction,
                                            invokation_timeout=self._invokation_timeout, 
                                            execution_timeout=self._execution_timeout)
    
    def noblock_set(self, value : typing.Any) -> None:
        return self._zmq_client.send_instruction(self._write_instruction, dict(value=value), 
                                            invokation_timeout=self._invokation_timeout, 
                                            execution_timeout=self._execution_timeout)
    
    def oneway_set(self, value : typing.Any) -> None:
        self._zmq_client.send_instruction(self._write_instruction, dict(value=value), 
                                            invokation_timeout=self._invokation_timeout, 
                                            execution_timeout=self._execution_timeout)
  


class _Event:
    
    __slots__ = ['_zmq_client', '_name', '_obj_name', '_unique_identifier', '_socket_address', '_callbacks', '_serialization_specific',
                    '_serializer', '_subscribed', '_thread', '_thread_callbacks', '_event_consumer', '_logger', '_deserialize']
    # event subscription
    # Dont add class doc otherwise __doc__ in slots will conflict with class variable

    def __init__(self, client : SyncZMQClient, name : str, obj_name : str, unique_identifier : str, socket : str, 
                    serialization_specific : bool = False, serializer : BaseSerializer = None, logger : logging.Logger = None) -> None:
        self._zmq_client = client
        self._name = name
        self._obj_name = obj_name
        self._unique_identifier = unique_identifier
        self._socket_address = socket
        self._serialization_specific = serialization_specific
        self._callbacks = None 
        self._serializer = serializer
        self._logger = logger 
        self._subscribed = False
        self._deserialize = True

    def add_callbacks(self, callbacks : typing.Union[typing.List[typing.Callable], typing.Callable]) -> None:
        if not self._callbacks:
            self._callbacks = [] 
        if isinstance(callbacks, list):
            self._callbacks.extend(callbacks)
        else:
            self._callbacks.append(callbacks)

    def subscribe(self, callbacks : typing.Union[typing.List[typing.Callable], typing.Callable], 
                    thread_callbacks : bool = False, deserialize : bool = True):
        self._event_consumer = EventConsumer(
                                    'zmq-' + self._unique_identifier if self._serialization_specific else self._unique_identifier, 
                                    self._socket_address, f"{self._name}|RPCEvent|{uuid.uuid4()}", b'PROXY',
                                    zmq_serializer=self._serializer, logger=self._logger
                                )
        self.add_callbacks(callbacks) 
        self._subscribed = True
        self._deserialize = deserialize
        self._thread_callbacks = thread_callbacks
        self._thread = threading.Thread(target=self.listen)
        self._thread.start()

    def listen(self):
        while self._subscribed:
            try:
                data = self._event_consumer.receive(deserialize=self._deserialize)
                if data == 'INTERRUPT':
                    break
                for cb in self._callbacks: 
                    if not self._thread_callbacks:
                        cb(data)
                    else: 
                        threading.Thread(target=cb, args=(data,)).start()
            except Exception as ex:
                warnings.warn(f"Uncaught exception from {self._name} event - {str(ex)}", 
                                category=RuntimeWarning)
        try:
            self._event_consumer.exit()
        except:
            pass
       

    def unsubscribe(self, join_thread : bool = True):
        self._subscribed = False
        self._event_consumer.interrupt()
        if join_thread:
            self._thread.join()
            
            



__allowed_attribute_types__ = (_Property, _RemoteMethod, _Event)
__WRAPPER_ASSIGNMENTS__ =  ('__name__', '__qualname__', '__doc__')

def _add_method(client_obj : ObjectProxy, method : _RemoteMethod, func_info : ZMQResource) -> None:
    if not func_info.top_owner:
        return 
        raise RuntimeError("logic error")
    for dunder in __WRAPPER_ASSIGNMENTS__:
        if dunder == '__qualname__':
            info = '{}.{}'.format(client_obj.__class__.__name__, func_info.get_dunder_attr(dunder).split('.')[1])
        else:
            info = func_info.get_dunder_attr(dunder)
        setattr(method, dunder, info)
    client_obj.__setattr__(func_info.obj_name, method)

def _add_property(client_obj : ObjectProxy, property : _Property, property_info : ZMQResource) -> None:
    if not property_info.top_owner:
        return
        raise RuntimeError("logic error")
    for attr in ['__doc__', '__name__']: 
        # just to imitate _add_method logic
        setattr(property, attr, property_info.get_dunder_attr(attr))
    client_obj.__setattr__(property_info.obj_name, property)

def _add_event(client_obj : ObjectProxy, event : _Event, event_info : ServerSentEvent) -> None:
    setattr(client_obj, event_info.obj_name, event)
    


class ReplyNotArrivedError(Exception):
    pass 


__all__ = ['ObjectProxy']




import typing
import numpy
from types import FunctionType

from .parameterized import Parameterized, ParameterizedMetaclass
from .parameters import *



class NumpyArray(ClassSelector):
    """
    Parameter whose value is a numpy array.
    """

    def __init__(self, default=None, doc : typing.Union[str, None] = None, 
            constant : bool = False, readonly : bool = False, allow_None : bool = False,
            label : typing.Union[str, None] = None,  per_instance : bool = False, deep_copy : bool = False, 
            class_member : bool = False, fget : FunctionType = None, fset : FunctionType = None,
            precedence : float = None) -> None:
        
        super(NumpyArray, self).__init__(class_=numpy.ndarray, default=default,  doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, label=label, per_instance=per_instance, 
            deep_copy=deep_copy, class_member=class_member, fget=fget, fset=fset, precedence=precedence)

    @typing.overload
    def __get__(self, obj : typing.Union[Parameterized, typing.Any], 
                objtype: typing.Union[ParameterizedMetaclass, typing.Any]) -> numpy.ndarray:
        ...

    @classmethod
    def serialize(cls, value : typing.Union[numpy.ndarray, None]):
        if value is None:
            return None
        return value.tolist()

    @classmethod
    def deserialize(cls, value):
        if value == 'null' or value is None:
            return None
        return numpy.asarray(value)
    

from pandas import DataFrame as pdDFrame
     

class DataFrame(ClassSelector):
    """
    Parameter whose value is a pandas DataFrame.

    The structure of the DataFrame can be constrained by the rows and
    columns arguments:

    rows: If specified, may be a number or an integer bounds tuple to
    constrain the allowable number of rows.

    columns: If specified, may be a number, an integer bounds tuple, a
    list or a set. If the argument is numeric, constrains the number of
    columns using the same semantics as used for rows. If either a list
    or set of strings, the column names will be validated. If a set is
    used, the supplied DataFrame must contain the specified columns and
    if a list is given, the supplied DataFrame must contain exactly the
    same columns and in the same order and no other columns.
    """

    __slots__ = ['rows', 'columns', 'ordered']

    def __init__(self, default=None, rows=None, columns=None, ordered=None, **params):
        self.rows = rows
        self.columns = columns
        self.ordered = ordered
        super(DataFrame,self).__init__(pdDFrame, default=default, **params)
        self._validate(self.default)

    def _length_bounds_check(self, bounds, length, name):
        message = '{name} length {length} does not match declared bounds of {bounds}'
        if not isinstance(bounds, tuple):
            if (bounds != length):
                raise ValueError(message.format(name=name, length=length, bounds=bounds))
            else:
                return
        (lower, upper) = bounds
        failure = ((lower is not None and (length < lower))
                   or (upper is not None and length > upper))
        if failure:
            raise ValueError(message.format(name=name,length=length, bounds=bounds))

    def _validate(self, val):
        super(DataFrame, self)._validate(val)

        if isinstance(self.columns, set) and self.ordered is True:
            raise ValueError('Columns cannot be ordered when specified as a set')

        if self.allow_None and val is None:
            return

        if self.columns is None:
            pass
        elif (isinstance(self.columns, tuple) and len(self.columns)==2
              and all(isinstance(v, (type(None), numbers.Number)) for v in self.columns)): # Numeric bounds tuple
            self._length_bounds_check(self.columns, len(val.columns), 'Columns')
        elif isinstance(self.columns, (list, set)):
            self.ordered = isinstance(self.columns, list) if self.ordered is None else self.ordered
            difference = set(self.columns) - set([str(el) for el in val.columns])
            if difference:
                msg = 'Provided DataFrame columns {found} does not contain required columns {expected}'
                raise ValueError(msg.format(found=list(val.columns), expected=sorted(self.columns)))
        else:
            self._length_bounds_check(self.columns, len(val.columns), 'Column')

        if self.ordered:
            if list(val.columns) != list(self.columns):
                msg = 'Provided DataFrame columns {found} must exactly match {expected}'
                raise ValueError(msg.format(found=list(val.columns), expected=self.columns))

        if self.rows is not None:
            self._length_bounds_check(self.rows, len(val), 'Row')

    @classmethod
    def serialize(cls, value):
        if value is None:
            return 'null'
        return value.to_dict('records')

    @classmethod
    def deserialize(cls, value):
        if value == 'null':
            return None
        from pandas import DataFrame as pdDFrame
        return pdDFrame(value)
    


class Series(ClassSelector):
    """
    Parameter whose value is a pandas Series.

    The structure of the Series can be constrained by the rows argument
    which may be a number or an integer bounds tuple to constrain the
    allowable number of rows.
    """

    __slots__ = ['rows']

    def __init__(self, default=None, rows=None, allow_None=False, **params):
        from pandas import Series as pdSeries
        self.rows = rows
        super(Series,self).__init__(pdSeries, default=default, allow_None=allow_None,
                                    **params)
        self._validate(self.default)

    def _length_bounds_check(self, bounds, length, name):
        message = '{name} length {length} does not match declared bounds of {bounds}'
        if not isinstance(bounds, tuple):
            if (bounds != length):
                raise ValueError(message.format(name=name, length=length, bounds=bounds))
            else:
                return
        (lower, upper) = bounds
        failure = ((lower is not None and (length < lower))
                   or (upper is not None and length > upper))
        if failure:
            raise ValueError(message.format(name=name,length=length, bounds=bounds))

    def _validate(self, val):
        super(Series, self)._validate(val)

        if self.allow_None and val is None:
            return

        if self.rows is not None:
            self._length_bounds_check(self.rows, len(val), 'Row')



import textwrap
import typing
from contextlib import contextmanager

def wrap_error_text(text : str) -> str:
    # return T.wrap(text)
    #'\n'+'\n'.join([line.lstrip() 
    return textwrap.fill(
        text = textwrap.dedent(text).lstrip(),
        initial_indent='\n',
        expand_tabs=True,
        replace_whitespace= True,
    )

def raise_TypeError(message, parameter) -> typing.NoReturn:
    owner_str = ''
    if isinstance(parameter, Parameter):
        owner_str = f" Owner info : {parameter.owner}, parameter name : {parameter.name}."
    elif issubclass(parameter, Parameter):
        owner_str = ''
    raise TypeError(message + owner_str)

def raise_ValueError(message, parameter) -> typing.NoReturn:
    owner_str = ''
    if isinstance(parameter, Parameter):
        owner_str = f" Owner info : {parameter.owner}, parameter name : {parameter.name}."
    elif issubclass(parameter, Parameter):
        owner_str = ''
    raise ValueError(message + owner_str)


def get_iterable_printfriendly_repr(iterable):
    # This method can be called before __init__ has called
    # super's __init__, so there may not be any name set yet.
    items = []
    limiter = ']'
    length = 0
    for item in iterable:
        string = str(item)
        length += len(string)
        if length < 200:
            items.append(string)
        else:
            limiter = ', ...]'
            break
    items = '[' + ', '.join(items) + limiter
    return items


@contextmanager
def exceptions_summarized():
    """Useful utility for writing docs that need to show expected errors.
    Shows exception only, concisely, without a traceback.
    """
    try:
        yield
    except Exception:
        import sys
        etype, value, tb = sys.exc_info()
        print("{}: {}".format(etype.__name__,value), file=sys.stderr)

from .parameterized import Parameter

__all__ = ['wrap_error_text', 'raise_TypeError', 'raise_ValueError', 'get_iterable_printfriendly_repr']


"""
namespace for pure RPC applications to auto suggest the user. The following were the original names
given until the WoT schema was being generated. The docs then got totally mixed up with two types 
of names and therefore the WoT naming was taken as final. 
"""
import typing
from enum import Enum
from ..server import Thing as RemoteObject, action
from ..server.constants import USE_OBJECT_NAME, HTTP_METHODS, JSON


def remote_method(URL_path : str = USE_OBJECT_NAME, http_method : str = HTTP_METHODS.POST, 
            state : typing.Optional[typing.Union[str, Enum]] = None, argument_schema : typing.Optional[JSON] = None,
            return_value_schema : typing.Optional[JSON] = None, **kwargs) -> typing.Callable:
    return action(URL_path=URL_path, http_method=http_method, state=state, argument_schema=argument_schema,
                return_value_schema=return_value_schema, **kwargs)


import typing
from types import FunctionType, MethodType
from enum import Enum
import warnings

from ..param.parameterized import Parameter, ClassParameters, Parameterized, ParameterizedMetaclass
from .utils import issubklass, pep8_to_URL_path
from .dataklasses import RemoteResourceInfoValidator
from .constants import USE_OBJECT_NAME, HTTP_METHODS
from .events import Event, EventDispatcher



class Property(Parameter):
    """
    Initialize a new Property object and to get/set/delete an object/instance attribute. Please note the capital 'P' in 
    Property to differentiate from python's own ``property``. ``Property`` objects are similar to python ``property``
    but not a subclass of it due to limitations and redundancy.

    Parameters
    ----------

    default: None or corresponding to property type 
        The default value of the property. 

    doc: str, default empty
        docstring explaining what this property represents.

    constant: bool, default False
        if True, the Property value can be changed only once when ``allow_None`` is set to True. The
        value is otherwise constant on the ``Thing`` instance.

    readonly: bool, default False
        if True, the Property value cannot be changed by setting the attribute at the class or instance
        levels at all. Either the value is fetched always or a getter method is executed which may still generate dynamic
        values at each get/read operation.

    allow_None: bool, default False 
        if True, None is accepted as a valid value for this Property, in addition to any other values that are
        allowed. 

    URL_path: str, uses object name by default
        resource locator under which the attribute is accessible through HTTP. When not given, the variable name 
        is used and underscores are replaced with dash

    http_method: tuple, default ("GET", "PUT", "DELETE")
        http methods for read, write and delete respectively 

    observable: bool, default False
        set to True to receive change events. Supply a function if interested to evaluate on what conditions the change 
        event must be emitted. Default condition is a plain not-equal-to operator.

    state: str | Enum, default None
        state of state machine where property write can be executed

    db_persist: bool, default False
        if True, every write is stored in database and property value persists ``Thing`` instance destruction and creation. 
        The loaded value from database is written into the property at ``Thing.__post_init__``.
        set ``Thing.use_default_db`` to True, to avoid setting up a database or supply a ``db_config_file``.
    
    db_init: bool, default False
        if True, property's first value is loaded from database and written using setter. 
        Further writes are not written to database. if ``db_persist`` is True, this value is ignored. 

    db_commit: bool,
        if True, all write values are stored to database. The database value is not loaded at ``Thing.__post_init__()``. 
        if db_persist is True, this value is ignored. 

    fget: Callable, default None
        custom getter method, mandatory when setter method is also custom supplied.

    fset: Callable, default None
        custom setter method
    
    fdel: Callable, default None
        custom deleter method

    remote: bool, default True
        set to false to make the property local/not remotely accessible

    label: str, default extracted from object name
        optional text label to be used when this Property is shown in a listing. If no label is supplied, 
        the attribute name for this property in the owning ``Thing`` object is used.
    
    metadata: dict, default None
        store your own JSON compatible metadata for the property which gives useful (and modifiable) information 
        about the property. Properties operate using slots which means you cannot set foreign attributes on this object 
        normally. This metadata dictionary should overcome this limitation. 

    per_instance_descriptor: bool, default False 
        whether a separate Property instance will be created for every ``Thing`` instance. True by default.
        If False, all instances of a ```Thing``` class will share the same Property object, including all validation
        attributes (bounds, allow_None etc.). 

    deepcopy_default: bool, default False 
        controls whether the default value of this Property will be deepcopied when a ``Thing`` object is instantiated (if
        True), or if the single default value will be shared by all ``Thing`` instances (if False). For an immutable 
        Property value, it is best to leave deep_copy at the default of False. For a mutable Property value, 
        for example - lists and dictionaries, the default of False is also appropriate if you want all instances to share 
        the same value state, e.g. if they are each simply referring to a single global object like a singleton. 
        If instead each ``Thing`` should have its own independently mutable value, deep_copy should be set to
        True. This setting is similar to using ``field``'s ``default_factory`` in python dataclasses.

    class_member : bool, default False
        when True, property is set on ``Thing`` class instead of ``Thing`` instance. 

    precedence: float, default None
        a numeric value, usually in the range 0.0 to 1.0, which allows the order of Properties in a class to be defined in
        a listing or e.g. in GUI menus. A negative precedence indicates a property that should be hidden in such listings.

    """

    __slots__ = ['db_persist', 'db_init', 'db_commit', 'metadata', 'model', '_remote_info', 
                '_observable', '_observable_event_descriptor', 'fcomparator', '_old_value_internal_name']
    
    # RPC only init - no HTTP methods for those who dont like
    @typing.overload
    def __init__(self, default: typing.Any = None, *, doc : typing.Optional[str] = None, constant : bool = False, 
                readonly : bool = False, allow_None : bool = False, observable : bool = False, 
                state : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,
                db_persist : bool = False, db_init : bool = False, db_commit : bool = False,  remote : bool = True,
                class_member : bool = False, fget : typing.Optional[typing.Callable] = None, 
                fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None,
            ) -> None:
        ...

    @typing.overload
    def __init__(self, default: typing.Any = None, *, doc : typing.Optional[str] = None, constant : bool = False, 
                readonly : bool = False, allow_None : bool = False, URL_path : str = USE_OBJECT_NAME, 
                http_method : typing.Tuple[typing.Optional[str], typing.Optional[str]] = (HTTP_METHODS.GET, HTTP_METHODS.PUT), 
                observable : bool = False, state : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,
                db_persist : bool = False, db_init : bool = False, db_commit : bool = False, remote : bool = True, 
                class_member : bool = False, fget : typing.Optional[typing.Callable] = None, 
                fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None, 
                metadata : typing.Optional[typing.Dict] = None
            ) -> None:
        ...

    @typing.overload
    def __init__(self, default: typing.Any = None, *, doc : typing.Optional[str] = None, constant : bool = False, 
                readonly : bool = False, allow_None : bool = False, 
                URL_path : str = USE_OBJECT_NAME, 
                http_method : typing.Tuple[typing.Optional[str], typing.Optional[str], typing.Optional[str]] = 
                                                            (HTTP_METHODS.GET, HTTP_METHODS.PUT, HTTP_METHODS.DELETE), 
                observable : bool = False, change_comparator : typing.Optional[typing.Union[FunctionType, MethodType]] = None,
                state : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,
                db_persist : bool = False, db_init : bool = False, db_commit : bool = False, remote : bool = True,
                class_member : bool = False, fget : typing.Optional[typing.Callable] = None, 
                fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None, 
                fcomparator : typing.Optional[typing.Callable] = None, 
                deepcopy_default : bool = False, per_instance_descriptor : bool = False, 
                precedence : typing.Optional[float] = None, metadata : typing.Optional[typing.Dict] = None
            ) -> None:
        ...
 
    def __init__(self, default: typing.Any = None, *, 
                doc : typing.Optional[str] = None, constant : bool = False, 
                readonly : bool = False, allow_None : bool = False, label : typing.Optional[str] = None, 
                URL_path : str = USE_OBJECT_NAME, 
                http_method : typing.Tuple[typing.Optional[str], typing.Optional[str], typing.Optional[str]] = 
                                                            (HTTP_METHODS.GET, HTTP_METHODS.PUT, HTTP_METHODS.DELETE), 
                state : typing.Optional[typing.Union[typing.List, typing.Tuple, str, Enum]] = None,
                db_persist : bool = False, db_init : bool = False, db_commit : bool = False, 
                observable : bool = False, class_member : bool = False, model = None, 
                fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None, 
                fdel : typing.Optional[typing.Callable] = None, fcomparator : typing.Optional[typing.Callable] = None,  
                deepcopy_default : bool = False, per_instance_descriptor : bool = False, remote : bool = True, 
                precedence : typing.Optional[float] = None, metadata : typing.Optional[typing.Dict] = None
            ) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, allow_None=allow_None,
                    label=label, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default,
                    class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.db_persist = db_persist
        self.db_init    = db_init
        self.db_commit  = db_commit
        self.fcomparator = fcomparator
        self.metadata = metadata
        self._observable = observable
        self._observable_event_descriptor : Event = None
        self._remote_info = None
        if remote:
            self._remote_info = RemoteResourceInfoValidator(
                http_method=http_method,
                URL_path=URL_path,
                state=state,
                isproperty=True
            )
        self.model = None
        if model:
            self.model = wrap_plain_types_in_rootmodel(model)


    def __set_name__(self, owner: typing.Any, attrib_name: str) -> None:
        super().__set_name__(owner, attrib_name)
        self._old_value_internal_name = f'{self._internal_name}_old_value'
        if self._remote_info is not None:
            if self._remote_info.URL_path == USE_OBJECT_NAME:
                self._remote_info.URL_path = f'/{pep8_to_URL_path(self.name)}'
            elif not self._remote_info.URL_path.startswith('/'): 
                raise ValueError(f"URL_path should start with '/', please add '/' before '{self._remote_info.URL_path}'")
            self._remote_info.obj_name = self.name
        if self._observable:
            _observable_event_name = f'{self.name}_change_event'  
            # This is a descriptor object, so we need to set it on the owner class
            self._observable_event_descriptor = Event(
                        friendly_name=_observable_event_name,
                        URL_path=f'{self._remote_info.URL_path}/change-event',
                        doc=f"change event for {self.name}"
                    ) # type: Event
            self._observable_event_descriptor.__set_name__(owner, _observable_event_name)
            setattr(owner, _observable_event_name, self._observable_event_descriptor)
          

    def _push_change_event_if_needed(self, obj, value : typing.Any) -> None:
        """
        Pushes change event both on read and write if an event publisher object is available
        on the owning Thing.        
        """
        if self._observable and obj.event_publisher:
            event_dispatcher = getattr(obj, self._observable_event_descriptor._obj_name, None) # type: EventDispatcher
            old_value = obj.__dict__.get(self._old_value_internal_name, NotImplemented)
            obj.__dict__[self._old_value_internal_name] = value 
            if self.fcomparator:
                if issubklass(self.fcomparator):
                    if not self.fcomparator(self.owner, old_value, value):
                        return 
                elif not self.fcomparator(obj, old_value, value):
                    return
            elif not old_value != value:
                return 
            event_dispatcher.push(value)       


    def __get__(self, obj: Parameterized, objtype: ParameterizedMetaclass) -> typing.Any:
        read_value = super().__get__(obj, objtype)
        self._push_change_event_if_needed(obj, read_value)
        return read_value
    

    def _post_value_set(self, obj, value : typing.Any) -> None:
        if (self.db_persist or self.db_commit) and hasattr(obj, 'db_engine'):
            from .thing import Thing
            assert isinstance(obj, Thing), f"database property {self.name} bound to a non Thing, currently not supported"
            obj.db_engine.set_property(self, value)
        self._push_change_event_if_needed(obj, value)
        return super()._post_value_set(obj, value)
    
    
    def comparator(self, func : typing.Callable) -> typing.Callable:
        """
        Register a comparator method by using this as a decorator to decide when to push
        a change event.
        """
        self.fcomparator = func 
        return func
    

    
__property_info__ = [
                'allow_None' , 'class_member', 'db_init', 'db_persist', 
                'db_commit', 'deepcopy_default', 'per_instance_descriptor', 
                'state', 'precedence', 'constant', 'default'
                # 'scada_info', 'property_type' # descriptor related info is also necessary
            ]

   
class ClassProperties(ClassParameters):
    """
    Object that holds the namespace and implementation of Parameterized methods as well as any state that is not 
    in __slots__ or the Properties themselves.
    Exists at metaclass level (instantiated by the metaclass). Contains state specific to the class.
    """

    @property
    def db_persisting_objects(self):
        try:
            return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_persisting_remote_params')
        except AttributeError: 
            paramdict = self.remote_objects
            db_persisting_remote_params = {}
            for name, desc in paramdict.items():
                if desc.db_persist:
                    db_persisting_remote_params[name] = desc
            setattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_persisting_remote_params', db_persisting_remote_params)
        return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_persisting_remote_params')

    @property
    def db_init_objects(self) -> typing.Dict[str, Property]:
        try:
            return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_init_remote_params')
        except AttributeError: 
            paramdict = self.remote_objects
            init_load_params = {}
            for name, desc in paramdict.items():
                if desc.db_init or desc.db_persist:
                    init_load_params[name] = desc
            setattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_init_remote_params', init_load_params)
        return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_db_init_remote_params')
        
    @property
    def remote_objects(self) -> typing.Dict[str, Property]:
        try:
            return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_remote_params')
        except AttributeError: 
            paramdict = super().descriptors
            remote_params = {}
            for name, desc in paramdict.items():
                if isinstance(desc, Property):
                    remote_params[name] = desc
            setattr(self.owner_cls, f'_{self.owner_cls.__name__}_remote_params', remote_params)
        return getattr(self.owner_cls, f'_{self.owner_cls.__name__}_remote_params')

    def webgui_info(self, for_remote_params : typing.Union[Property, typing.Dict[str, Property], None] = None):
        info = {}
        if isinstance(for_remote_params, dict):
            objects = for_remote_params 
        elif isinstance(for_remote_params, Property):
            objects = { for_remote_params.name : for_remote_params } 
        else:
            objects = self.remote_objects
        for param in objects.values():
            state = param.__getstate__()
            info[param.name] = dict(
                python_type = param.__class__.__name__,
            )
            for field in __property_info__:
                info[param.name][field] = state.get(field, None) 
        return info 

    
try: 
    from pydantic import BaseModel, RootModel, create_model
    def wrap_plain_types_in_rootmodel(model : type) -> type["BaseModel"]:
        """
        Ensure a type is a subclass of BaseModel.

        If a `BaseModel` subclass is passed to this function, we will pass it
        through unchanged. Otherwise, we wrap the type in a RootModel.
        In the future, we may explicitly check that the argument is a type
        and not a model instance.
        """
        try:  # This needs to be a `try` as basic types are not classes
            assert issubclass(model, BaseModel)
            return model
        except (TypeError, AssertionError):
            return create_model(f"{model!r}", root=(model, ...), __base__=RootModel)
        except NameError:
            raise ImportError("pydantic is not installed, please install it to use this feature") from None
except ImportError:
    def wrap_plain_types_in_rootmodel(model : type) -> type:
        raise ImportError("pydantic is not installed, please install it to use this feature") from None
  

__all__ = [
    Property.__name__
]


import datetime
import logging
import unittest
import time
import os
from hololinked.client import ObjectProxy
from hololinked.server import action, Thing, global_config
from hololinked.server.properties import Number, String, Selector, List, Integer
from hololinked.server.database import BaseDB
try:
    from .utils import TestCase, TestRunner
    from .things import start_thing_forked
except ImportError:
    from utils import TestCase, TestRunner
    from things import start_thing_forked




class TestThing(Thing):

    number_prop = Number(doc="A fully editable number property")
    string_prop = String(default='hello', regex='^[a-z]+',
                        doc="A string property with a regex constraint to check value errors")
    int_prop = Integer(default=5, step=2, bounds=(0, 100),
                        doc="An integer property with step and bounds constraints to check RW")
    selector_prop = Selector(objects=['a', 'b', 'c', 1], default='a',
                        doc="A selector property to check RW")
    observable_list_prop = List(default=None, allow_None=True, observable=True,
                        doc="An observable list property to check observable events on write operations")
    observable_readonly_prop = Number(default=0, readonly=True, observable=True,
                        doc="An observable readonly property to check observable events on read operations")
    db_commit_number_prop = Number(default=0, db_commit=True,
                        doc="A fully editable number property to check commits to db on write operations")
    db_init_int_prop = Integer(default=1, db_init=True,
                        doc="An integer property to check initialization from db")
    db_persist_selector_prop = Selector(objects=['a', 'b', 'c', 1], default='a', db_persist=True,
                        doc="A selector property to check persistence to db on write operations")
    non_remote_number_prop = Number(default=5, remote=False,
                        doc="A non remote number property to check non-availability on client")

    @observable_readonly_prop.getter
    def get_observable_readonly_prop(self):
        if not hasattr(self, '_observable_readonly_prop'):
            self._observable_readonly_prop = 0
        self._observable_readonly_prop += 1
        return self._observable_readonly_prop


    @action()
    def print_props(self):
        print(f'number_prop: {self.number_prop}')
        print(f'string_prop: {self.string_prop}')
        print(f'int_prop: {self.int_prop}')
        print(f'selector_prop: {self.selector_prop}')
        print(f'observable_list_prop: {self.observable_list_prop}')
        print(f'observable_readonly_prop: {self.observable_readonly_prop}')
        print(f'db_commit_number_prop: {self.db_commit_number_prop}')
        print(f'db_init_int_prop: {self.db_init_int_prop}')
        print(f'db_persist_selctor_prop: {self.db_persist_selector_prop}')
        print(f'non_remote_number_prop: {self.non_remote_number_prop}')




class TestProperty(TestCase):

    @classmethod
    def setUpClass(self):
        print("test property")
        self.thing_cls = TestThing
        start_thing_forked(self.thing_cls, instance_name='test-property',
                                    log_level=logging.WARN)
        self.thing_client = ObjectProxy('test-property') # type: TestThing

    @classmethod
    def tearDownClass(self):
        print("tear down test property")
        self.thing_client.exit()


    def test_1_client_api(self):
        # Test read
        self.assertEqual(self.thing_client.number_prop, 0)
        # Test write
        self.thing_client.string_prop = 'world'
        self.assertEqual(self.thing_client.string_prop, 'world')
        # Test exception propagation to client
        with self.assertRaises(ValueError):
            self.thing_client.string_prop = 'WORLD'
        with self.assertRaises(TypeError):
            self.thing_client.int_prop = '5'
        # Test non remote prop (non-)availability on client
        with self.assertRaises(AttributeError):
            self.thing_client.non_remote_number_prop


    def test_2_RW_multiple_properties(self):
        # Test partial list of read write properties
        self.thing_client.write_multiple_properties(
                number_prop=15,
                string_prop='foobar'
            )
        self.assertEqual(self.thing_client.number_prop, 15)
        self.assertEqual(self.thing_client.string_prop, 'foobar')
        # check prop that was not set in multiple properties
        self.assertEqual(self.thing_client.int_prop, 5)

        self.thing_client.selector_prop = 'b'
        self.thing_client.number_prop = -15
        props = self.thing_client.read_multiple_properties(names=['selector_prop', 'int_prop',
                                                    'number_prop', 'string_prop'])
        self.assertEqual(props['selector_prop'], 'b')
        self.assertEqual(props['int_prop'], 5)
        self.assertEqual(props['number_prop'], -15)
        self.assertEqual(props['string_prop'], 'foobar')


    def test_3_observability(self):
        # req 1 - observable events come due to writing a property
        propective_values = [
            [1, 2, 3, 4, 5],
            ['a', 'b', 'c', 'd', 'e'],
            [1, 'a', 2, 'b', 3]
        ]
        result = []
        attempt = 0
        def cb(value):
            nonlocal attempt, result
            self.assertEqual(value, propective_values[attempt])
            result.append(value)
            attempt += 1

        self.thing_client.subscribe_event('observable_list_prop_change_event', cb)
        time.sleep(3)
        # Calm down for event publisher to connect fully as there is no handshake for events
        for value in propective_values:
            self.thing_client.observable_list_prop = value

        for i in range(20):
            if attempt == len(propective_values):
                break
            # wait for the callback to be called
            time.sleep(0.1)
        self.thing_client.unsubscribe_event('observable_list_prop_change_event')

        self.assertEqual(result, propective_values)

        # req 2 - observable events come due to reading a property
        propective_values = [1, 2, 3, 4, 5]
        result = []
        attempt = 0
        def cb(value):
            nonlocal attempt, result
            self.assertEqual(value, propective_values[attempt])
            result.append(value)
            attempt += 1

        self.thing_client.subscribe_event('observable_readonly_prop_change_event', cb)
        time.sleep(3)
        # Calm down for event publisher to connect fully as there is no handshake for events
        for _ in propective_values:
            self.thing_client.observable_readonly_prop

        for i in range(20):
            if attempt == len(propective_values):
                break
            # wait for the callback to be called
            time.sleep(0.1)

        self.thing_client.unsubscribe_event('observable_readonly_prop_change_event')
        self.assertEqual(result, propective_values)


    def test_4_db_operations(self):
        # remove old file path first
        file_path = f'{BaseDB.get_temp_dir_for_class_name(TestThing.__name__)}/test-db-operations.db'
        try:
            os.remove(file_path)
        except (OSError, FileNotFoundError):
            pass
        self.assertTrue(not os.path.exists(file_path))
    	
        # test db commit property
        thing = TestThing(instance_name='test-db-operations', use_default_db=True, log_level=logging.WARN)
        self.assertEqual(thing.db_commit_number_prop, 0) # 0 is default just for reference
        thing.db_commit_number_prop = 100
        self.assertEqual(thing.db_commit_number_prop, 100)
        self.assertEqual(thing.db_engine.get_property('db_commit_number_prop'), 100)

        # test db persist property
        self.assertEqual(thing.db_persist_selector_prop, 'a') # a is default just for reference
        thing.db_persist_selector_prop = 'c'
        self.assertEqual(thing.db_persist_selector_prop, 'c')
        self.assertEqual(thing.db_engine.get_property('db_persist_selector_prop'), 'c')

        # test db init property
        self.assertEqual(thing.db_init_int_prop, 1) # 1 is default just for reference
        thing.db_init_int_prop = 50
        self.assertEqual(thing.db_init_int_prop, 50)
        self.assertNotEqual(thing.db_engine.get_property('db_init_int_prop'), 50)
        self.assertEqual(thing.db_engine.get_property('db_init_int_prop'), TestThing.db_init_int_prop.default)
        del thing

        # delete thing and reload from database 
        thing = TestThing(instance_name='test-db-operations', use_default_db=True, log_level=logging.WARN)
        self.assertEqual(thing.db_init_int_prop, TestThing.db_init_int_prop.default)
        self.assertEqual(thing.db_persist_selector_prop, 'c')
        self.assertNotEqual(thing.db_commit_number_prop, 100)
        self.assertEqual(thing.db_commit_number_prop, TestThing.db_commit_number_prop.default)

        # check db init prop with a different value in database apart from default
        thing.db_engine.set_property('db_init_int_prop', 101)
        del thing
        thing = TestThing(instance_name='test-db-operations', use_default_db=True, log_level=logging.WARN)
        self.assertEqual(thing.db_init_int_prop, 101)


if __name__ == '__main__':
    unittest.main(testRunner=TestRunner())



import setuptools

# read the contents of your README file
from pathlib import Path
long_description = (Path(__file__).parent/"README.md").read_text()


setuptools.setup(
    name="hololinked",
    version="0.2.7",
    author="Vignesh Vaidyanathan",
    author_email="vignesh.vaidyanathan@hololinked.dev",
    description="A ZMQ-based Object Oriented RPC tool-kit for instrument control/data acquisition or controlling generic python objects.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://hololinked.readthedocs.io/en/latest/index.html",
    packages=[
        'hololinked',
        'hololinked.server',
        'hololinked.rpc',
        'hololinked.client',
        'hololinked.param'    
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Manufacturing", 
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Education",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Browsers",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: System :: Hardware",
        "Development Status :: 4 - Beta"
    ],    
    python_requires='>=3.11',
    install_requires=[
        "argon2-cffi>=23.0.0",
        "ifaddr>=0.2.0",
        "msgspec>=0.18.6",
        "pyzmq>=25.1.0",
        "SQLAlchemy>=2.0.21",
        "SQLAlchemy_Utils>=0.41.1",
        "tornado>=6.3.3",
        "jsonschema>=4.22.0"
    ],
    license="BSD-3-Clause",
    license_files=('license.txt', 'licenses/param-LICENSE.txt', 'licenses/pyro-LICENSE.txt'),
    keywords=["data-acquisition", "zmq-rpc", "SCADA/IoT", "Web of Things"]
)
 


from __future__ import annotations

from enum import Enum
from pydantic import BaseModel, Field, ConfigDict, TypeAdapter, ValidationError
from pydantic.json_schema import GenerateJsonSchema
from pydantic._internal._core_utils import is_core_schema, CoreSchemaOrField
from typing import Optional, Sequence, Union, Any, Mapping, List, Dict
from .serializers import JSONSerializer

JSONSchema = dict[str, Any]  # A type to represent JSONSchema

AnyUri = str
Description = str
Descriptions = Optional[Dict[str, str]]
Title = str
Titles = Optional[Dict[str, str]]
Security = Union[List[str], str]
Scopes = Union[List[str], str]
TypeDeclaration = Union[str, List[str]]


class Type(Enum):
    boolean = "boolean"
    integer = "integer"
    number = "number"
    string = "string"
    object = "object"
    array = "array"
    null = "null"


class DataSchema(BaseModel):

    field_type: Optional[TypeDeclaration] = Field(None, alias="@type")
    description: Optional[Description] = None
    title: Optional[Title] = None
    descriptions: Optional[Descriptions] = None
    titles: Optional[Titles] = None
    writeOnly: Optional[bool] = None
    readOnly: Optional[bool] = None
    oneOf: Optional[list[DataSchema]] = None
    unit: Optional[str] = None
    enum: Optional[list] = None
    # enum was `Field(None, min_length=1, unique_items=True)` but this failed with
    # generic models
    format: Optional[str] = None
    const: Optional[Any] = None
    default: Optional[Any] = None
    type: Optional[Type] = None
    # The fields below should be empty unless type==Type.array
    items: Optional[Union[DataSchema, List[DataSchema]]] = None
    maxItems: Optional[int] = Field(None, ge=0)
    minItems: Optional[int] = Field(None, ge=0)
    # The fields below should be empty unless type==Type.number or Type.integer
    minimum: Optional[Union[int, float]] = None
    maximum: Optional[Union[int, float]] = None
    exclusiveMinimum: Optional[Union[int, float]] = None
    exclusiveMaximum: Optional[Union[int, float]] = None
    multipleOf: Optional[Union[int, float]] = None
    # The fields below should be empty unless type==Type.object
    properties: Optional[Mapping[str, DataSchema]] = None
    required: Optional[list[str]] = None
    # The fields below should be empty unless type==Type.string
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    pattern: Optional[str] = None
    contentEncoding: Optional[str] = None
    contentMediaType: Optional[str] = None

    model_config = ConfigDict(extra="forbid")



def is_a_reference(d: JSONSchema) -> bool:
    """Return True if a JSONSchema dict is a reference

    JSON Schema references are one-element dictionaries with
    a single key, `$ref`.  `pydantic` sometimes breaks this
    rule and so I don't check that it's a single key.
    """
    return "$ref" in d


def look_up_reference(reference: str, d: JSONSchema) -> JSONSchema:
    """Look up a reference in a JSONSchema

    This first asserts the reference is local (i.e. starts with #
    so it's relative to the current file), then looks up
    each path component in turn.
    """
    if not reference.startswith("#/"):
        raise NotImplementedError(
            "Built-in resolver can only dereference internal JSON references "
            "(i.e. starting with #)."
        )
    try:
        resolved: JSONSchema = d
        for key in reference[2:].split("/"):
            resolved = resolved[key]
        return resolved
    except KeyError as ke:
        raise KeyError(
            f"The JSON reference {reference} was not found in the schema "
            f"(original error {ke})."
        )


def is_an_object(d: JSONSchema) -> bool:
    """Determine whether a JSON schema dict is an object"""
    return "type" in d and d["type"] == "object"


def convert_object(d: JSONSchema) -> JSONSchema:
    """Convert an object from JSONSchema to Thing Description"""
    out: JSONSchema = d.copy()
    # AdditionalProperties is not supported by Thing Description, and it is ambiguous
    # whether this implies it's false or absent. I will, for now, ignore it, so we
    # delete the key below.
    if "additionalProperties" in out:
        del out["additionalProperties"]
    return out


def convert_anyof(d: JSONSchema) -> JSONSchema:
    """Convert the anyof key to oneof

    JSONSchema makes a distinction between "anyof" and "oneof", where the former
    means "any of these fields can be present" and the latter means "exactly one
    of these fields must be present". Thing Description does not have this
    distinction, so we convert anyof to oneof.
    """
    if "anyOf" not in d:
        return d
    out: JSONSchema = d.copy()
    out["oneOf"] = out["anyOf"]
    del out["anyOf"]
    return out


def convert_prefixitems(d: JSONSchema) -> JSONSchema:
    """Convert the prefixitems key to items

    JSONSchema 2019 (as used by thing description) used
    `items` with a list of values in the same way that JSONSchema
    now uses `prefixitems`.

    JSONSchema 2020 uses `items` to mean the same as `additionalItems`
    in JSONSchema 2019 - but Thing Description doesn't support the
    `additionalItems` keyword. This will result in us overwriting
    additional items, and we raise a ValueError if that happens.

    This behaviour may be relaxed in the future.
    """
    if "prefixItems" not in d:
        return d
    out: JSONSchema = d.copy()
    if "items" in out:
        raise ValueError(f"Overwrote the `items` key on {out}.")
    out["items"] = out["prefixItems"]
    del out["prefixItems"]
    return out


def convert_additionalproperties(d: JSONSchema) -> JSONSchema:
    """Move additionalProperties into properties, or remove it"""
    if "additionalProperties" not in d:
        return d
    out: JSONSchema = d.copy()
    if "properties" in out and "additionalProperties" not in out["properties"]:
        out["properties"]["additionalProperties"] = out["additionalProperties"]
    del out["additionalProperties"]
    return out


def check_recursion(depth: int, limit: int):
    """Check the recursion count is less than the limit"""
    if depth > limit:
        raise ValueError(
            f"Recursion depth of {limit} exceeded - perhaps there is a circular "
            "reference?"
        )


def jsonschema_to_dataschema(
    d: JSONSchema,
    root_schema: Optional[JSONSchema] = None,
    recursion_depth: int = 0,
    recursion_limit: int = 99,
) -> JSONSchema:
    """remove references and change field formats

    JSONSchema allows schemas to be replaced with `{"$ref": "#/path/to/schema"}`.
    Thing Description does not allow this. `dereference_jsonschema_dict` takes a
    `dict` representation of a JSON Schema document, and replaces all the
    references with the appropriate chunk of the file.

    JSONSchema can represent `Union` types using the `anyOf` keyword, which is
    called `oneOf` by Thing Description.  It's possible to achieve the same thing
    in the specific case of array elements, by setting `items` to a list of
    `DataSchema` objects. This function does not yet do that conversion.

    This generates a copy of the document, to avoid messing up `pydantic`'s cache.
    """
    root_schema = root_schema or d
    check_recursion(recursion_depth, recursion_limit)
    # JSONSchema references are one-element dictionaries, with a single key called $ref
    while is_a_reference(d):
        d = look_up_reference(d["$ref"], root_schema)
        recursion_depth += 1
        check_recursion(recursion_depth, recursion_limit)

    if is_an_object(d):
        d = convert_object(d)
    d = convert_anyof(d)
    d = convert_prefixitems(d)
    d = convert_additionalproperties(d)

    # After checking the object isn't a reference, we now recursively check
    # sub-dictionaries and dereference those if necessary. This could be done with a
    # comprehension, but I am prioritising readability over speed. This code is run when
    # generating the TD, not in time-critical situations.
    rkwargs: dict[str, Any] = {
        "root_schema": root_schema,
        "recursion_depth": recursion_depth + 1,
        "recursion_limit": recursion_limit,
    }
    output: JSONSchema = {}
    for k, v in d.items():
        if isinstance(v, dict):
            # Any items that are Mappings (i.e. sub-dictionaries) must be recursed into
            output[k] = jsonschema_to_dataschema(v, **rkwargs)
        elif isinstance(v, Sequence) and len(v) > 0 and isinstance(v[0], Mapping):
            # We can also have lists of mappings (i.e. Array[DataSchema]), so we
            # recurse into these.
            output[k] = [jsonschema_to_dataschema(item, **rkwargs) for item in v]
        else:
            output[k] = v
    return output


def type_to_dataschema(t: Union[type, BaseModel], **kwargs) -> DataSchema:
    """Convert a Python type to a Thing Description DataSchema

    This makes use of pydantic's `schema_of` function to create a
    json schema, then applies some fixes to make a DataSchema
    as per the Thing Description (because Thing Description is
    almost but not quite compatible with JSONSchema).

    Additional keyword arguments are added to the DataSchema,
    and will override the fields generated from the type that
    is passed in. Typically you'll want to use this for the
    `title` field.
    """
    if isinstance(t, BaseModel):
        json_schema = t.model_json_schema()
    else:
        json_schema = TypeAdapter(t).json_schema()
    schema_dict = jsonschema_to_dataschema(json_schema)
    # Definitions of referenced ($ref) schemas are put in a
    # key called "definitions" or "$defs" by pydantic. We should delete this.
    # TODO: find a cleaner way to do this
    # This shouldn't be a severe problem: we will fail with a
    # validation error if other junk is left in the schema.
    for k in ["definitions", "$defs"]:
        if k in schema_dict:
            del schema_dict[k]
    schema_dict.update(kwargs)
    try:
        return DataSchema(**schema_dict)
    except ValidationError as ve:
        print(
            "Error while constructing DataSchema from the "
            "following dictionary:\n"
            + JSONSerializer().dumps(schema_dict, indent=2)
            + "Before conversion, the JSONSchema was:\n"
            + JSONSerializer().dumps(json_schema, indent=2)
        )
        raise ve
    

class GenerateJsonSchemaWithoutDefaultTitles(GenerateJsonSchema):
    """Drops autogenerated titles from JSON Schema"""
    
    # https://stackoverflow.com/questions/78679812/pydantic-v2-to-json-schema-translation-how-to-suppress-autogeneration-of-title
    def field_title_should_be_set(self, schema: CoreSchemaOrField) -> bool:
        return_value = super().field_title_should_be_set(schema)
        if return_value and is_core_schema(schema):
            return False
        return return_value
